from copy import deepcopy

from NEMO.constants import CHAR_FIELD_MEDIUM_LENGTH
from NEMO.models import Project
from NEMO.serializers import ModelSerializer, ProjectSerializer
from NEMO.utilities import export_format_datetime
from NEMO.views.api import ModelViewSet, ProjectViewSet, date_filters, datetime_filters, key_filters
from NEMO.views.api_billing import BillingFilterForm
from django import forms
from django.db.models import Q
from drf_excel.mixins import XLSXFileMixin
from rest_flex_fields.serializers import FlexFieldsSerializerMixin
from rest_framework import status
from rest_framework.fields import BooleanField, CharField, DateTimeField, DecimalField, IntegerField, empty
from rest_framework.response import Response
from rest_framework.serializers import Serializer
from rest_framework.viewsets import GenericViewSet

from NEMO_billing.customization import BillingCustomization
from NEMO_billing.invoices.customization import InvoiceCustomization
from NEMO_billing.invoices.models import InvoiceConfiguration, InvoicePayment, ProjectBillingDetails
from NEMO_billing.invoices.processors import invoice_data_processor_class as data_processor
from NEMO_billing.invoices.views.usage import augment_with_invoice_items


class NewBillingFilterForm(BillingFilterForm):
    include_no_charge_projects = forms.BooleanField(initial=False, required=False)

    def get_include_no_charge_projects(self):
        return self.cleaned_data["include_no_charge_projects"]


class ProjectBillingDetailsSerializer(FlexFieldsSerializerMixin, ModelSerializer):
    class Meta:
        model = ProjectBillingDetails
        exclude = ("project",)


class ProjectWithDetailsSerializer(ProjectSerializer):
    projectbillingdetails = ProjectBillingDetailsSerializer(required=False)

    class Meta(ProjectSerializer.Meta):
        ProjectSerializer.Meta.expandable_fields.update(
            {
                "category": (
                    "NEMO_billing.rates.api.RateCategorySerializer",
                    {"source": "projectbillingdetails.category"},
                ),
                "institution": (
                    "NEMO_billing.api.InstitutionSerializer",
                    {"source": "projectbillingdetails.institution"},
                ),
                "department": ("NEMO_billing.api.DepartmentSerializer", {"source": "projectbillingdetails.department"}),
            }
        )

    def get_fields(self):
        fields = super().get_fields()
        detail_fields = fields.pop("projectbillingdetails", {})
        if detail_fields:
            for key, value in detail_fields.fields.items():
                if key != "id":
                    # reset the source to details
                    value.source = "projectbillingdetails." + value.source
                    value.source_attrs = value.source.split(".")
                    fields[key] = value
        return fields

    def validate(self, attrs):
        attributes_data = dict(attrs)
        details_attrs = attributes_data.pop("projectbillingdetails", {})
        super().validate(attributes_data)
        project_details = get_billing_details(self.instance)
        for details_attr, details_value in details_attrs.items():
            setattr(project_details, details_attr, details_value)
        ProjectBillingDetailsSerializer().full_clean(project_details, exclude=["project"])
        return attrs

    def update(self, instance, validated_data) -> Project:
        data = deepcopy(validated_data)
        details_data = data.pop("projectbillingdetails", {})
        project_instance = super().update(instance, data)
        project_details = get_billing_details(project_instance)
        ProjectBillingDetailsSerializer().update(project_details, details_data)
        return project_instance

    def create(self, validated_data) -> Project:
        data = deepcopy(validated_data)
        details_data = data.pop("projectbillingdetails", {})
        project_instance = super().create(data)
        details_data["project"] = project_instance
        ProjectBillingDetailsSerializer().create(details_data)
        return project_instance


class ProjectWithDetailsViewSet(ProjectViewSet):
    serializer_class = ProjectWithDetailsSerializer


class InvoicePaymentSerializer(ModelSerializer):
    class Meta:
        model = InvoicePayment
        fields = "__all__"


class InvoicePaymentViewSet(ModelViewSet):
    filename = "invoice_payments"
    queryset = InvoicePayment.objects.all()
    serializer_class = InvoicePaymentSerializer
    filterset_fields = {
        "id": key_filters,
        "payment_received": date_filters,
        "payment_processed": date_filters,
        "created_date": datetime_filters,
        "updated_date": datetime_filters,
        "invoice_id": key_filters,
    }


class BillingDataSerializer(Serializer):
    item_type = CharField(source="item_type.display_name", read_only=True)
    item_id = IntegerField(source="item.id", read_only=True)
    core_facility = CharField(max_length=CHAR_FIELD_MEDIUM_LENGTH, read_only=True)
    name = CharField(max_length=CHAR_FIELD_MEDIUM_LENGTH, read_only=True)
    account = CharField(source="project.account.name", read_only=True)
    account_id = IntegerField(source="project.account.id", read_only=True)
    project = CharField(source="project.name", read_only=True)
    project_id = IntegerField(source="project.id", read_only=True)
    institution = CharField(source="project.projectbillingdetails.institution.name", read_only=True, allow_null=True)
    institution_id = IntegerField(
        source="project.projectbillingdetails.institution.id", read_only=True, allow_null=True
    )
    department = CharField(source="project.projectbillingdetails.department.name", read_only=True, allow_null=True)
    department_id = IntegerField(source="project.projectbillingdetails.department.id", read_only=True, allow_null=True)
    application = CharField(source="project.application_identifier", read_only=True)
    reference_po = CharField(source="project.application_identifier", read_only=True)
    user = CharField(source="user.username", read_only=True)
    user_fullname = CharField(source="user.get_name", read_only=True)
    proxy_user = CharField(source="proxy_user.username", read_only=True, allow_null=True)
    proxy_user_fullname = CharField(source="proxy_user.get_name", read_only=True, allow_null=True)
    start = DateTimeField(read_only=True)
    end = DateTimeField(read_only=True)
    quantity = DecimalField(read_only=True, decimal_places=2, max_digits=8)
    rate = CharField(source="billable_rate", read_only=True)
    rate_category = CharField(source="rate.category", read_only=True, allow_null=True)
    rate_time_name = CharField(source="rate.time.name", read_only=True, allow_null=True)
    rate_time_id = IntegerField(source="rate.time.id", read_only=True, allow_null=True)
    unit_rate = DecimalField(source="rate.amount", read_only=True, decimal_places=2, max_digits=14, allow_null=True)
    unit_type = CharField(read_only=True, allow_null=True)
    unit_quantity = DecimalField(read_only=True, decimal_places=2, max_digits=14, allow_null=True)
    validated = BooleanField(read_only=True)
    validated_by = CharField(source="validated_by.username", read_only=True, allow_null=True)
    waived = BooleanField(read_only=True)
    waived_by = CharField(read_only=True, source="waived_by.username", allow_null=True)
    waived_on = DateTimeField(read_only=True)
    amount = DecimalField(source="invoiced_amount", read_only=True, decimal_places=2, max_digits=14)
    discount_amount = DecimalField(source="invoiced_discount", read_only=True, decimal_places=2, max_digits=14)
    pending_amount = DecimalField(source="amount", read_only=True, decimal_places=2, max_digits=14)
    merged_amount = DecimalField(label="amount", read_only=True, decimal_places=2, max_digits=14)

    def __init__(self, instance=None, data=empty, **kwargs):
        super().__init__(instance, data, **kwargs)
        pending_vs_final = BillingCustomization.get_bool("billing_usage_show_pending_vs_final")
        if not pending_vs_final:
            self.fields["amount"] = self.fields["merged_amount"]
            del self.fields["pending_amount"]
        del self.fields["merged_amount"]

    def update(self, instance, validated_data):
        pass

    def create(self, validated_data):
        pass

    class Meta:
        fields = "__all__"


class BillingDataViewSet(XLSXFileMixin, GenericViewSet):
    serializer_class = BillingDataSerializer

    def list(self, request, *args, **kwargs):
        billing_form = NewBillingFilterForm(self.request.GET)
        if not billing_form.is_valid():
            return Response(status=status.HTTP_400_BAD_REQUEST, data=billing_form.errors)
        try:
            queryset = self.get_queryset()
        except Exception as e:
            return Response(status=status.HTTP_400_BAD_REQUEST, data=str(e))
        serializer = self.serializer_class(queryset, many=True)
        return Response(serializer.data)

    def check_permissions(self, request):
        if not request or not request.user.has_perm("NEMO.use_billing_api"):
            self.permission_denied(request)

    def get_queryset(self):
        billing_form = NewBillingFilterForm(self.request.GET)
        billing_form.full_clean()

        queryset = Q()
        if not billing_form.get_include_no_charge_projects():
            queryset &= Q(project__projectbillingdetails__no_charge=False)
        if InvoiceCustomization.get_bool("invoice_skip_inactive_projects"):
            queryset &= Q(project__active=True)
        if InvoiceCustomization.get_bool("invoice_skip_inactive_accounts"):
            queryset &= Q(project__account__active=True)
        start, end = billing_form.get_start_date(), billing_form.get_end_date()
        if billing_form.get_account_id():
            queryset &= Q(project__account_id=billing_form.get_account_id())
        if billing_form.get_account_name():
            queryset &= Q(project__account__name=billing_form.get_account_name())
        if billing_form.get_project_id():
            queryset &= Q(project_id=billing_form.get_project_id())
        if billing_form.get_project_name():
            queryset &= Q(project__name=billing_form.get_project_name())
        if billing_form.get_application_name():
            queryset &= Q(project__application_identifier=billing_form.get_application_name())
        user_filter, customer_filter, trainee_filter = queryset, queryset, queryset
        if billing_form.get_username():
            user_filter = user_filter & Q(user__username=billing_form.get_username())
            customer_filter = customer_filter & Q(customer__username=billing_form.get_username())
            trainee_filter = trainee_filter & Q(trainee__username=billing_form.get_username())

        config = InvoiceConfiguration.first_or_default()
        billables = data_processor.get_billable_items(start, end, config, customer_filter, user_filter, trainee_filter)
        augment_with_invoice_items(billables)
        billables.sort(key=lambda x: x.start, reverse=True)
        return billables

    def get_filename(self, *args, **kwargs):
        return f"billing-{export_format_datetime()}.xlsx"


def get_billing_details(project: Project):
    try:
        project_details = project.projectbillingdetails
    except (ProjectBillingDetails.DoesNotExist, AttributeError):
        project_details = ProjectBillingDetails(project=project)
    return project_details



# Generated by Django 4.2.11 on 2024-08-20 16:06

import re

import django.core.validators
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("invoices", "0012_invoiceconfiguration_invoice_title"),
        ("NEMO", "0086_adjustmentrequest_new_quantity"),
        ("NEMO_billing", "0005_version_2_5_6"),
    ]

    operations = [
        migrations.CreateModel(
            name="ProjectBillingHardCap",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("enabled", models.BooleanField(default=True)),
                ("start_date", models.DateField()),
                ("end_date", models.DateField()),
                ("amount", models.DecimalField(decimal_places=2, max_digits=14)),
                (
                    "charge_types",
                    models.CharField(
                        help_text="List of charge types that will count towards this CAP",
                        max_length=100,
                        validators=[
                            django.core.validators.RegexValidator(
                                re.compile("^\\d+(?:,\\d+)*\\Z"),
                                code="invalid",
                                message="Enter only digits separated by commas.",
                            )
                        ],
                    ),
                ),
                (
                    "configuration",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="invoices.invoiceconfiguration",
                    ),
                ),
                ("project", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="NEMO.project")),
            ],
            options={
                "ordering": ["-start_date"],
            },
        ),
    ]



from NEMO.models import Account, Area, Project, Tool, User

from NEMO_billing.invoices.models import ProjectBillingDetails
from NEMO_billing.rates.models import RateCategory


def basic_data(set_rate_category=False) -> (User, Project, Tool, Area):
    test_user = User.objects.create(
        username="test", first_name="Test", last_name="Test", is_staff=False, badge_number=1
    )
    test_project = Project.objects.create(name="Test Project", account=Account.objects.create(name="Test Account"))
    if set_rate_category:
        academia = RateCategory.objects.create(name="Academia")
        industry = RateCategory.objects.create(name="Industry")
        test_project.projectbillingdetails = ProjectBillingDetails.objects.create(
            project=test_project, category=academia
        )
    else:
        test_project.projectbillingdetails = ProjectBillingDetails.objects.create(project=test_project)
    test_user.projects.add(test_project)
    tool = Tool.objects.create(name="tool")
    area = Area.objects.create(name="Cleanroom")
    return test_user, test_project, tool, area



from __future__ import annotations

from datetime import MAXYEAR, MINYEAR, date, datetime
from decimal import Decimal
from typing import Dict, List, Optional, Tuple

from NEMO.constants import CHAR_FIELD_SMALL_LENGTH
from NEMO.models import Account, BaseModel, User
from NEMO.typing import QuerySetType
from NEMO.utilities import RecurrenceFrequency, format_datetime, get_recurring_rule
from dateutil.rrule import rrule
from django.core.exceptions import ValidationError
from django.core.validators import MaxValueValidator, MinValueValidator, validate_comma_separated_integer_list
from django.db import IntegrityError, models, transaction
from django.db.models import Q, UniqueConstraint
from django.db.models.signals import m2m_changed
from django.dispatch import receiver

from NEMO_billing.invoices.models import BillableItemType
from NEMO_billing.models import CoreFacility
from NEMO_billing.rates.models import RateCategory
from NEMO_billing.utilities import Months


class CAPDiscountConfiguration(BaseModel):
    rate_category = models.ForeignKey(
        RateCategory, help_text="The rate category this CAP applies to", on_delete=models.CASCADE
    )
    core_facilities = models.ManyToManyField(
        CoreFacility,
        blank=True,
        help_text="The core facility(ies) this CAP applies to. If multiple are selected, the CAP is shared between facilities",
    )
    charge_types = models.CharField(
        validators=[validate_comma_separated_integer_list],
        max_length=CHAR_FIELD_SMALL_LENGTH,
        help_text="List of charge type this CAP applies to",
    )
    start_on_first_charge = models.BooleanField(
        default=True, help_text="Automatically set CAP start date to the month of the first eligible charge"
    )
    split_by_user = models.BooleanField(
        default=True,
        help_text="Check this box to have a separate CAP amount per each user/account. If unchecked, the CAP will be per account only (all user charges count towards the same CAP)",
    )
    # Recurring schedule
    start_month = models.PositiveIntegerField(
        null=True, blank=True, choices=Months.choices, help_text="Start month of the CAP discount"
    )
    start_year = models.PositiveIntegerField(
        null=True,
        blank=True,
        help_text="Start year of the CAP discount",
        validators=[MinValueValidator(1900), MaxValueValidator(MAXYEAR)],
    )
    reset_interval = models.PositiveIntegerField(
        default=1,
        validators=[MinValueValidator(1)],
        help_text="Recurring interval, i.e. every 5 days.",
    )
    reset_frequency = models.PositiveIntegerField(
        choices=[
            (RecurrenceFrequency.MONTHLY.index, RecurrenceFrequency.MONTHLY.display_value),
            (RecurrenceFrequency.YEARLY.index, RecurrenceFrequency.YEARLY.display_value),
        ],
        help_text="The CAP reset frequency.",
    )
    creation_time = models.DateTimeField(auto_now_add=True, null=True, blank=True)

    @property
    def start(self) -> date:
        if self.start_year and self.start_month:
            return date(self.start_year, self.start_month, 1)

    @start.setter
    def start(self, value: date):
        if value:
            self.start_month, self.start_year = value.month, value.year
        else:
            self.start_month, self.start_year = None, None

    @property
    def billable_charge_types(self) -> List[BillableItemType]:
        return [BillableItemType(int(value)) for value in self.charge_types.split(",") if value]

    @property
    def core_facility_names(self) -> Tuple[str]:
        return tuple(cf.name for cf in self.core_facilities.all())

    @property
    def get_reset_frequency_enum(self):
        return RecurrenceFrequency(self.reset_frequency)

    def clean(self):
        errors = {}
        # We need to check ids otherwise it will throw a RelatedObjectDoesNotExist if they are not set even if required
        if self.start_month and not self.start_year or self.start_year and not self.start_month:
            errors["start_year"] = "Both year/month are required together"
            errors["start_month"] = "Both year/month are required together"
        if not self.start and not self.start_on_first_charge:
            errors["start_year"] = "Year is required when `Start on first charge` is unchecked."
            errors["start_month"] = "Month is required when `Start on first charge` is unchecked."
        if self.start and self.start_on_first_charge:
            errors["start_on_first_charge"] = "You entered a start date, remove it or unchecked this box."
        if errors:
            raise ValidationError(errors)

    def get_recurrence_interval_display(self) -> str:
        if not self.reset_frequency:
            return ""
        interval = f"{self.reset_interval} " if self.reset_interval != 1 else ""
        f_enum = self.get_reset_frequency_enum
        frequency = f"{f_enum.display_text}s" if self.reset_interval != 1 else f_enum.display_text
        return f"Every {interval}{frequency}"

    @transaction.atomic
    def get_or_create_cap_discount(self, account: Account, username: Optional[str], start: date):
        try:
            # Cap discount exists, return it
            return CAPDiscount.objects.get(configuration=self, account=account, user__username=username)
        except CAPDiscount.DoesNotExist:
            # Cap discount doesn't exist, create it
            user = User.objects.get(username=username) if username else None
            new_cap_discount = CAPDiscount(configuration=self, account=account, user=user)
            cap_discount_tiers = self.initialize_cap_discount_and_get_tiers(new_cap_discount)
            if self.start_on_first_charge:
                assert start is not None
                new_cap_discount.start = start
            new_cap_discount.save()
            # Now that we saved our new cap discount, save related tiers and initialize amounts
            new_cap_discount.save_related_cap_tiers(cap_discount_tiers)
            assert new_cap_discount.start is not None
            CAPDiscountAmount.objects.create(
                cap_discount=new_cap_discount, year=start.year, month=start.month, start=0, end=0
            )
            # Also create first amount
            return new_cap_discount

    def initialize_cap_discount_and_get_tiers(self, cap_discount: CAPDiscount) -> List[CAPDiscountTier]:
        # Initialize cap discount and return cap discount tiers to be saved later
        if self.start and not self.start_on_first_charge:
            cap_discount.start = self.start
        cap_discount.reset_interval = self.reset_interval
        cap_discount.reset_frequency = self.reset_frequency
        cap_discount.charge_types = self.charge_types
        return self.copy_cap_discount_tiers(cap_discount)

    def copy_cap_discount_tiers(self, cap_discount: CAPDiscount) -> List[CAPDiscountTier]:
        # Copy the discount tiers by creating new ones for the cap discount
        tiers = []
        for cap_discount_tier in self.capdiscounttier_set.all():
            tiers.append(
                CAPDiscountTier(
                    cap_discount=cap_discount, amount=cap_discount_tier.amount, discount=cap_discount_tier.discount
                )
            )
        return tiers

    def __str__(self):
        facilities = f"{','.join(self.core_facility_names)}/" if self.core_facilities.exists() else ""
        return f"CAP Discount for {facilities}{self.rate_category.name}"


# We have no choice but to check for uniqueness here, since many to many fields cannot be used
# in validate_unique or in clean methods
@receiver(m2m_changed, sender=CAPDiscountConfiguration.core_facilities.through)
def validate_unique_cap_config(sender, instance: CAPDiscountConfiguration, action, **kwargs):
    if action in ["post_add", "post_remove", "post_clear"]:
        matching_config: QuerySetType[CAPDiscountConfiguration] = CAPDiscountConfiguration.objects.filter(
            rate_category=instance.rate_category
        ).exclude(pk=instance.pk)
        match: Optional[CAPDiscountConfiguration] = None
        if not instance.core_facilities.exists():
            match = matching_config.filter(core_facilities__isnull=True).first()
        else:
            for cap_config in matching_config:
                cap_config_facilities = set(cap_config.core_facilities.all())
                intersection = cap_config_facilities.intersection(instance.core_facilities.all())
                if intersection:
                    match = cap_config
                    break
        if match:
            raise IntegrityError(
                f"There is already a CAP Configuration for {instance.rate_category} and the selected core facility(ies): {match}"
            )


class CAPDiscount(BaseModel):
    configuration = models.ForeignKey(
        CAPDiscountConfiguration, help_text="The configuration for this CAP", on_delete=models.CASCADE
    )
    user = models.ForeignKey(
        User, blank=True, null=True, help_text="The user this CAP applies to", on_delete=models.CASCADE
    )
    account = models.ForeignKey(Account, help_text="The account this CAP applies to", on_delete=models.CASCADE)
    charge_types = models.CharField(
        validators=[validate_comma_separated_integer_list],
        max_length=CHAR_FIELD_SMALL_LENGTH,
        help_text="List of charge type this CAP applies to",
    )
    # Recurring schedule
    start_month = models.PositiveIntegerField(
        null=True, blank=True, choices=Months.choices, help_text="Start month of the CAP discount"
    )
    start_year = models.PositiveIntegerField(
        null=True,
        blank=True,
        help_text="Start year of the CAP discount",
        validators=[MinValueValidator(1900), MaxValueValidator(MAXYEAR)],
    )
    next_reset_override_month = models.PositiveIntegerField(
        null=True, blank=True, choices=Months.choices, help_text="Temporary one-time reset override month"
    )
    next_reset_override_year = models.PositiveIntegerField(
        null=True,
        blank=True,
        help_text="Temporary one-time reset override year",
        validators=[MinValueValidator(1900), MaxValueValidator(MAXYEAR)],
    )
    reset_interval = models.PositiveIntegerField(
        default=1,
        validators=[MinValueValidator(1)],
        help_text="Recurring interval, i.e. every 5 days.",
    )
    reset_frequency = models.PositiveIntegerField(
        choices=[
            (RecurrenceFrequency.MONTHLY.index, RecurrenceFrequency.MONTHLY.display_value),
            (RecurrenceFrequency.YEARLY.index, RecurrenceFrequency.YEARLY.display_value),
        ],
        help_text="The CAP reset frequency.",
    )
    creation_time = models.DateTimeField(auto_now_add=True, null=True, blank=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Temporary variables used for discount calculations
        self.tmp_current_amount: Optional[Decimal] = None
        # Dict of project_id -> core facility -> Decimal
        self.tmp_project_discounts: Dict[int, Dict[str, Decimal]] = {}

    @property
    def start(self) -> date:
        if self.start_year and self.start_month:
            return date(self.start_year, self.start_month, 1)

    @start.setter
    def start(self, value: date):
        if value:
            self.start_month, self.start_year = value.month, value.year
        else:
            self.start_month, self.start_year = None, None

    @property
    def next_reset_override(self) -> date:
        if self.next_reset_override_year and self.next_reset_override_month:
            return date(self.next_reset_override_year, self.next_reset_override_month, 1)

    @next_reset_override.setter
    def next_reset_override(self, value: date):
        if value:
            self.next_reset_override_month, self.next_reset_override_year = value.month, value.year
        else:
            self.next_reset_override_month, self.next_reset_override_year = None, None

    @property
    def billable_charge_types(self) -> List[BillableItemType]:
        return [BillableItemType(int(value)) for value in self.charge_types.split(",") if value]

    @property
    def get_reset_frequency_enum(self):
        return RecurrenceFrequency(self.reset_frequency)

    def get_recurrence(self) -> rrule:
        if self.start and self.reset_frequency:
            return get_recurring_rule(self.start, self.get_reset_frequency_enum, interval=self.reset_interval)

    def get_recurrence_interval_display(self) -> str:
        if not self.start or not self.reset_frequency:
            return ""
        interval = f"{self.reset_interval} " if self.reset_interval != 1 else ""
        f_enum = self.get_reset_frequency_enum
        frequency = f"{f_enum.display_text}s" if self.reset_interval != 1 else f_enum.display_text
        return f"Every {interval}{frequency}"

    def get_recurrence_display(self) -> str:
        display = ""
        if self.start and self.reset_frequency:
            return f"{self.get_recurrence_interval_display()}, starting {format_datetime(self.start, 'F Y')}"
        return display

    def next_reset_date_after(self, dt: date, inc=False) -> datetime:
        if self.next_reset_override:
            return datetime(self.next_reset_override.year, self.next_reset_override.month, self.next_reset_override.day)
        dt_datetime = datetime(dt.year, dt.month, dt.day)
        recurrence = self.get_recurrence()
        return recurrence.after(dt_datetime, inc=inc) if recurrence else None

    def next_reset(self, inc=False) -> datetime:
        return self.next_reset_date_after(date.today(), inc=inc)

    def last_reset(self) -> datetime:
        recurrence = self.get_recurrence()
        return recurrence.before(datetime.now(), inc=False) if recurrence else None

    @transaction.atomic
    def reset(self):
        # Reinitialize based on the linked configuration, and blank out reset override
        cap_discount_tiers = self.configuration.initialize_cap_discount_and_get_tiers(self)
        # Save the new tiers
        self.save_related_cap_tiers(cap_discount_tiers)
        if self.next_reset_override and self.configuration.start_on_first_charge:
            # We are resetting the start date/month to the override to keep the recurrence correct
            self.start = self.next_reset_override
        self.next_reset_override = None
        self.save()

    def latest_amount(self) -> CAPDiscountAmount:
        return self.capdiscountamount_set.latest("year", "month")

    def earliest_amount(self) -> CAPDiscountAmount:
        return self.capdiscountamount_set.earliest("year", "month")

    def save_related_cap_tiers(self, cap_discount_tiers):
        for cap_discount_tier in cap_discount_tiers:
            cap_discount_tier.save()
        self.capdiscounttier_set.set(cap_discount_tiers)

    def clean(self):
        errors = {}
        # We need to check ids otherwise it will throw a RelatedObjectDoesNotExist if they are not set even if required
        if self.account_id and self.user_id and not self.user.projects.filter(account=self.account).exists():
            errors["account"] = "This user doesn't belong to this account"
        if self.start_month and not self.start_year or self.start_year and not self.start_month:
            errors["start_year"] = "Both year/month are required together"
            errors["start_month"] = "Both year/month are required together"
        if (
            self.next_reset_override_month
            and not self.next_reset_override_year
            or self.next_reset_override_year
            and not self.next_reset_override_month
        ):
            errors["next_reset_override_year"] = "Both year/month are required together"
            errors["next_reset_override_month"] = "Both year/month are required together"
        this_month = date(date.today().year, date.today().month, 1)
        if self.next_reset_override is not None and self.next_reset_override < this_month:
            errors["next_reset_override_year"] = "The override date cannot be in the past."
            errors["next_reset_override_month"] = "The override date cannot be in the past."
        if errors:
            raise ValidationError(errors)

    def estimated_charges_since_last_update(self) -> Decimal:
        from NEMO_billing.cap_discount.processors import CAPDiscountInvoiceDataProcessor

        return CAPDiscountInvoiceDataProcessor().estimated_charges_since_last_update(self)

    def current_level_reached(self) -> CAPDiscountTier:
        total_current_charges = self.latest_amount().end + self.estimated_charges_since_last_update()
        for tier in self.capdiscounttier_set.all().reverse():
            if total_current_charges > tier.amount:
                return tier

    def __str__(self):
        username = f"/{self.user.username}" if self.user else ""
        facilities = (
            f"{','.join(self.configuration.core_facility_names)}/"
            if self.configuration.core_facilities.exists()
            else ""
        )
        return f"CAP Discount for {facilities}{self.account}/{self.configuration.rate_category.name}{username}"

    class Meta:
        constraints = [
            UniqueConstraint(
                fields=["configuration", "account", "user"],
                name="%(app_label)s_%(class)s_unique_config_account_user",
            ),
            UniqueConstraint(
                fields=["configuration", "account"],
                condition=Q(user=None),
                name="%(app_label)s_%(class)s_unique_config_account_null_user",
            ),
        ]
        ordering = ["-creation_time", "configuration", "account", "user"]


class CAPDiscountTier(BaseModel):
    cap_discount_configuration = models.ForeignKey(
        CAPDiscountConfiguration, null=True, blank=True, on_delete=models.CASCADE
    )
    cap_discount = models.ForeignKey(CAPDiscount, null=True, blank=True, on_delete=models.CASCADE)
    amount = models.DecimalField(
        decimal_places=2,
        max_digits=14,
        help_text="Amount after which the discount will be applied",
        validators=[MinValueValidator(1)],
    )
    discount = models.DecimalField(
        decimal_places=3,
        max_digits=6,
        help_text="Discount in percent. Ex 20.5%",
        validators=[MinValueValidator(0), MaxValueValidator(100)],
    )

    def discount_amount(self) -> Decimal:
        return -self.discount / Decimal(100)

    def clean(self):
        if not self.cap_discount and not self.cap_discount_configuration:
            raise ValidationError("One of cap discount or cap discount configuration needs to be set")
        if self.cap_discount and self.cap_discount_configuration:
            raise ValidationError("Discount tiers cannot be linked to both cap discount and cap discount configuration")

    class Meta:
        ordering = ["amount"]

    def __str__(self):
        return f"> {str(self.amount).rstrip('0').rstrip('.')} -> {str(self.discount).rstrip('0').rstrip('.')} %"


class CAPDiscountAmount(BaseModel):
    cap_discount = models.ForeignKey(CAPDiscount, on_delete=models.CASCADE)
    month = models.PositiveIntegerField(choices=Months.choices)
    year = models.PositiveIntegerField(
        validators=[MinValueValidator(MINYEAR), MaxValueValidator(MAXYEAR)],
    )
    start = models.DecimalField(decimal_places=2, max_digits=14)
    end = models.DecimalField(decimal_places=2, max_digits=14)

    def new_charges(self):
        return (self.end or 0) - (self.start or 0)

    @property
    def amount_date(self) -> date:
        if self.year and self.month:
            return date(self.year, self.month, 1)

    @amount_date.setter
    def amount_date(self, value: date):
        if value:
            self.month, self.year = value.month, value.year
        else:
            self.month, self.year = None, None

    def __str__(self):
        return f"CAP #{self.cap_discount.id}, {self.get_month_display()} {self.year}, {self.start} -> {self.end}"

    class Meta:
        ordering = ["-year", "-month"]
        unique_together = ["cap_discount", "month", "year"]



# Generated by Django 4.2.19 on 2025-02-21 03:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("prepayments", "0002_projectprepaymentdetail_overdraft_amount_allowed"),
    ]

    operations = [
        migrations.AlterField(
            model_name="fundtype",
            name="name",
            field=models.CharField(help_text="The unique name for this item", max_length=255, unique=True),
        ),
    ]



from NEMO.models import Area, Consumable, Project, Tool

from NEMO_billing.exceptions import BillingException
from NEMO_billing.invoices.models import Invoice, InvoiceDetailItem
from NEMO_billing.rates.models import RateCategory, RateType


class NoRateSetException(BillingException):
    def __init__(
        self,
        rate_type: RateType,
        category: RateCategory = None,
        tool: Tool = None,
        area: Area = None,
        consumable: Consumable = None,
    ):
        self.rate_type = rate_type
        self.category = category
        self.tool = tool
        self.area = area
        self.consumable = consumable
        for_category = f" for category: {category}" if category else ""
        for_item = (
            f" for: {tool if tool else area if area else consumable if consumable else ''}"
            if tool or area or consumable
            else ""
        )
        msg = f"No {self.rate_type.get_type_display()} rate is set{for_item}{for_category}"
        super().__init__(msg)


class NoProjectCategorySetException(BillingException):
    def __init__(self, rate_type: RateType, project: Project):
        self.rate_type = rate_type
        self.project = project
        msg = f"{self.rate_type.get_type_display()} is category specific but no category is set on project {project}"
        super().__init__(msg)


class NoProjectDetailsSetException(BillingException):
    def __init__(self, project: Project):
        self.project = project
        msg = f"There are no project details set for project {project}"
        super().__init__(msg)


class InvoiceAlreadyExistException(BillingException):
    def __init__(self, invoice: Invoice):
        self.invoice = invoice
        msg = f"An invoice ({invoice.invoice_number}) already exist for this project for this date range. Void it to be able to generate it again"
        super().__init__(msg)


class InvoiceItemsNotInFacilityException(BillingException):
    def __init__(self, item: InvoiceDetailItem):
        self.item = item
        item_type_display = item.get_item_type_display().replace("_", " ")
        msg = f"Error generating invoice. A {item_type_display}: {item.name} for user {item.user} is not part of any core facilities"
        super().__init__(msg)


class InvoiceGenerationException(BillingException):
    def __init__(self, invoice: Invoice, renderer, e: Exception):
        self.invoice = invoice
        self.renderer = renderer
        msg = f"Error rendering invoice {invoice.invoice_number} with {renderer.__class__.__name__} for project {invoice.project_details.project_name}: {str(e)}."
        super().__init__(msg)



PLUGIN_DEPENDENCY_CHECK = False
DEBUG = True
AUTH_USER_MODEL = "NEMO.User"

DATETIME_FORMAT = "l, F jS, Y @ g:i A"
DATE_FORMAT = "m/d/Y"
TIME_FORMAT = "g:i A"
DATETIME_INPUT_FORMATS = ["%m/%d/%Y %I:%M %p"]
DATE_INPUT_FORMATS = ["%m/%d/%Y"]
TIME_INPUT_FORMATS = ["%I:%M %p"]

USE_I18N = False
USE_L10N = False
USE_TZ = True

ROOT_URLCONF = "NEMO.urls"
ALLOW_CONDITIONAL_URLS = True

STATIC_URL = "/static/"
MEDIA_URL = "/media/"

LOGIN_URL = "login"
LOGIN_REDIRECT_URL = "login"
AUTHENTICATION_BACKENDS = ["NEMO.views.authentication.NginxKerberosAuthorizationHeaderAuthenticationBackend"]

SECRET_KEY = "test"

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.admin",
    "django.contrib.humanize",
    "NEMO_billing",
    "NEMO_billing.rates",
    "NEMO_billing.invoices",
    "NEMO_billing.cap_discount",
    "NEMO_billing.prepayments",
    "NEMO_billing.quotes",
    "NEMO.apps.kiosk",
    "NEMO.apps.area_access",
    "NEMO",
    "rest_framework",
    "django_filters",
    "mptt",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "django.middleware.common.BrokenLinkEmailsMiddleware",
    "NEMO.middleware.DeviceDetectionMiddleware",
    "NEMO.middleware.RemoteUserAuthenticationMiddleware",
]

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "NEMO.context_processors.hide_logout_button",  # Add a 'request context processor' in order to figure out whether to display the logout button. If the site is configured to use the LDAP authentication backend then we want to provide a logoff button (in the menu bar). Otherwise the Kerberos authentication backend is used and no logoff button is necessary.
                "NEMO.context_processors.base_context",  # Base context processor
                "django.contrib.auth.context_processors.auth",
                "django.template.context_processors.debug",
                "django.template.context_processors.media",
                "django.template.context_processors.static",
                "django.template.context_processors.request",
                "django.template.context_processors.tz",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

TIME_ZONE = "America/New_York"

REST_FRAMEWORK = {
    "DEFAULT_PERMISSION_CLASSES": ("NEMO.permissions.BillingAPI",),
    "DEFAULT_FILTER_BACKENDS": ("django_filters.rest_framework.DjangoFilterBackend",),
}

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "./test_nemo.db",
    }
}



from NEMO.urls import router, sort_urls
from django.urls import path

from NEMO_billing.prepayments import api
from NEMO_billing.prepayments.views import prepayments

# Rest API URLs
router.register(r"billing/funds", api.FundViewSet)
router.register(r"billing/fund_types", api.FundTypeViewSet)
router.register(r"billing/project_prepayments", api.ProjectPrepaymentDetailViewSet)
router.registry.sort(key=sort_urls)

urlpatterns = [
    path("usage_project_prepayments/", prepayments.usage_project_prepayments, name="usage_project_prepayments"),
    path("prepaid_project_status/", prepayments.prepaid_project_status, name="prepaid_project_status"),
]



import datetime
from typing import List, Optional

from NEMO.models import (
    Consumable,
    Project,
    Reservation,
    StaffCharge,
    Tool,
    User,
)
from NEMO.policy import NEMOPolicy

from NEMO_billing.exceptions import ChargeTypeNotAllowedForProjectException
from NEMO_billing.invoices.models import BillableItemType
from NEMO_billing.prepayments.exceptions import (
    ProjectFundsFacilityNotAllowedException,
    ProjectFundsNotSetException,
    ProjectInsufficientFundsException,
)


# This policy class is not meant to be instantiated directly. It will be used by the generic NEMOBillingPolicy
# See NEMOBillingPolicy for details on how we restrict charge types
class PrepaymentPolicy(NEMOPolicy):
    def check_to_enable_tool(self, tool: Tool, operator: User, user: User, project, staff_charge, remote_work=False):
        billable_type = BillableItemType.STAFF_CHARGE if (staff_charge or remote_work) else BillableItemType.TOOL_USAGE
        self.check_project_prepayment_charge(project, billable_type)

    def check_to_save_reservation(
        self,
        cancelled_reservation: Optional[Reservation],
        new_reservation: Reservation,
        user_creating_reservation: User,
        explicit_policy_override: bool,
    ) -> (List[str], bool):
        charge_type = BillableItemType.TOOL_USAGE if new_reservation.tool else BillableItemType.AREA_ACCESS
        self.check_project_prepayment_charge(new_reservation.project, charge_type)

    def check_billing_to_project(self, project: Project, user: User, item=None, charge=None, *args, **kwargs):
        if hasattr(project, "projectprepaymentdetail"):
            if isinstance(item, StaffCharge):
                self.check_project_prepayment_charge(project, BillableItemType.STAFF_CHARGE)
            self.check_prepayment_allowed_for_core_facility(project, item)
            self.check_prepayment_status_for_project(project)

    def check_prepayment_allowed_for_core_facility(self, project: Project, item):
        only_facilities = project.projectprepaymentdetail.only_core_facilities.all()
        check_item = not isinstance(item, Consumable) or item.core_facility
        if only_facilities and check_item and item.core_facility not in only_facilities:
            raise ProjectFundsFacilityNotAllowedException(project, only_facilities)

    def check_prepayment_status_for_project(self, project: Project):
        # Project has to have prepayments
        if not project.projectprepaymentdetail.fund_set.exists():
            raise ProjectFundsNotSetException(project)
        else:
            prepayment = project.projectprepaymentdetail
            date_to_check = datetime.date.today()
            # This function will raise ProjectFundsExpiredException and ProjectFundsInactiveException
            # At least one prepayment has to be active and not expired
            if not prepayment.active_funds(date_to_check).exists():
                raise ProjectInsufficientFundsException(project)
            else:
                # check total available funds (check since last balance update)
                # we have to check each month separately and calculate available funds and charges
                # otherwise we would miss expiring funds or funds not active yet
                # This method will throw an exception if any month period between dates has insufficient funds
                prepayment.get_prepayment_info(until=date_to_check, raise_exception=True)

    def check_project_prepayment_charge(self, project: Project, billable_item_type: BillableItemType):
        if (
            hasattr(project, "projectprepaymentdetail")
            and billable_item_type not in project.projectprepaymentdetail.billable_charge_types
        ):
            raise ChargeTypeNotAllowedForProjectException(project, billable_item_type)



from copy import deepcopy
from typing import Optional

from NEMO.admin import AreaAdmin, AreaAdminForm, ConsumableAdmin, StaffChargeAdmin, ToolAdmin, ToolAdminForm
from NEMO.models import Area, Consumable, StaffCharge, Tool
from django import forms
from django.conf import settings
from django.contrib import admin
from django.contrib.admin import register, widgets
from django.contrib.admin.widgets import FilteredSelectMultiple
from django.core.exceptions import ValidationError
from django.db.models import Q

from NEMO_billing.customization import BillingCustomization
from NEMO_billing.invoices.models import BillableItemType
from NEMO_billing.models import (
    CoreFacility,
    CoreRelationship,
    CustomCharge,
    Department,
    Institution,
    InstitutionType,
    ProjectBillingHardCap,
)
from NEMO_billing.templatetags.billing_tags import cap_discount_installed
from NEMO_billing.utilities import IntMultipleChoiceField, hide_form_field

STATES = [
    ("AL", "Alabama"),
    ("AK", "Alaska"),
    ("AS", "American Samoa"),
    ("AZ", "Arizona"),
    ("AR", "Arkansas"),
    ("CA", "California"),
    ("CO", "Colorado"),
    ("CT", "Connecticut"),
    ("DE", "Delaware"),
    ("DC", "District of Columbia"),
    ("FL", "Florida"),
    ("GA", "Georgia"),
    ("GU", "Guam"),
    ("HI", "Hawaii"),
    ("ID", "Idaho"),
    ("IL", "Illinois"),
    ("IN", "Indiana"),
    ("IA", "Iowa"),
    ("KS", "Kansas"),
    ("KY", "Kentucky"),
    ("LA", "Louisiana"),
    ("ME", "Maine"),
    ("MD", "Maryland"),
    ("MA", "Massachusetts"),
    ("MI", "Michigan"),
    ("MN", "Minnesota"),
    ("MS", "Mississippi"),
    ("MO", "Missouri"),
    ("MT", "Montana"),
    ("NE", "Nebraska"),
    ("NV", "Nevada"),
    ("NH", "New Hampshire"),
    ("NJ", "New Jersey"),
    ("NM", "New Mexico"),
    ("NY", "New York"),
    ("NC", "North Carolina"),
    ("ND", "North Dakota"),
    ("MP", "Northern Mariana Islands"),
    ("OH", "Ohio"),
    ("OK", "Oklahoma"),
    ("OR", "Oregon"),
    ("PA", "Pennsylvania"),
    ("PR", "Puerto Rico"),
    ("RI", "Rhode Island"),
    ("SC", "South Carolina"),
    ("SD", "South Dakota"),
    ("TN", "Tennessee"),
    ("TX", "Texas"),
    ("UT", "Utah"),
    ("VT", "Vermont"),
    ("VI", "Virgin Islands"),
    ("VA", "Virginia"),
    ("WA", "Washington"),
    ("WV", "West Virginia"),
    ("WI", "Wisconsin"),
    ("WY", "Wyoming"),
]


def changed_or_added(change, original_set, current_set):
    # If the model object is being changed then we can get the list of previous members.
    if change:
        original_members = set(original_set)
    else:  # The model object is being created (instead of changed) so we can assume there are no members (initially).
        original_members = set()
    current_members = set(current_set)
    added_members = []
    removed_members = []

    # Log membership changes if they occurred.
    symmetric_difference = original_members ^ current_members
    if symmetric_difference:
        if change:  # the members have changed, so find out what was added and removed...
            # We can see the previous members of the object model by looking it up
            # in the database because the member list hasn't been committed yet.
            added_members = set(current_members) - set(original_members)
            removed_members = set(original_members) - set(current_members)

        else:  # a model object is being created (instead of changed) so we can assume all the members are new...
            added_members = current_set
    return added_members, removed_members


def save_all_core_facility_relationships(
    current_items, original_items, core_facility: CoreFacility, field: str, change
):
    added_items, removed_items = changed_or_added(change, original_items, current_items)
    for item in added_items:
        save_or_delete_core_facility(item, core_facility, field)
    for item in removed_items:
        save_or_delete_core_facility(item, None, field)


def save_or_delete_core_facility(obj, core_facility: Optional[CoreFacility], field):
    has_core_relationship = hasattr(obj, "core_rel")
    if core_facility:
        if not has_core_relationship:
            obj.core_rel = CoreRelationship()
        obj.core_rel.core_facility = core_facility
        setattr(obj.core_rel, field, obj)
        obj.core_rel.save()
    elif not core_facility and has_core_relationship:
        obj.core_rel.delete()


class ListTextWidget(forms.TextInput):
    def __init__(self, data_list, name, values_only=False, *args, **kwargs):
        super(ListTextWidget, self).__init__(*args, **kwargs)
        self._name = name
        self._list = data_list
        self._values_only = values_only
        self.attrs.update({"list": "list__{}".format(self._name)})

    def render(self, name, value, attrs=None, renderer=None):
        text_html = super(ListTextWidget, self).render(name, value, attrs=attrs)
        data_list = '<datalist id="list__{}">'.format(self._name)
        for item in self._list:
            value = item[0] if not self._values_only else item[1]
            data_list += '<option value="{}">{}</option>'.format(value, item[1])
        data_list += "</datalist>"

        return text_html + data_list


class InstitutionForm(forms.ModelForm):
    class Meta:
        model = Institution
        fields = "__all__"
        widgets = {"state": ListTextWidget(data_list=STATES, name="state_list", values_only=True)}


@register(Institution)
class InstitutionAdmin(admin.ModelAdmin):
    list_display = ["name", "institution_type", "state", "get_country_display", "zip_code"]
    form = InstitutionForm


class CoreFacilityAdminForm(forms.ModelForm):
    class Meta:
        model = CoreFacility
        fields = "__all__"

    core_facility_tools = forms.ModelMultipleChoiceField(
        queryset=Tool.objects.all(),
        required=False,
        widget=FilteredSelectMultiple(verbose_name="Core facility tools", is_stacked=False),
    )
    core_facility_areas = forms.ModelMultipleChoiceField(
        queryset=Area.objects.all(),
        required=False,
        widget=FilteredSelectMultiple(verbose_name="Core facility areas", is_stacked=False),
    )
    core_facility_consumables = forms.ModelMultipleChoiceField(
        queryset=Consumable.objects.all(),
        required=False,
        widget=FilteredSelectMultiple(verbose_name="Core facility consumable", is_stacked=False),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # We are filtering out already set tools, areas and consumables
        no_facility_filter = Q(core_rel__isnull=True)
        if "external_id" in self.fields:
            self.fields["external_id"].label = BillingCustomization.get("billing_core_facility_external_id_name")

        # Exclude children tools since their core facility is their parent's
        if "core_facility_tools" in self.fields:
            tool_filter = Q()
            if self.instance.pk:
                tool_filter = Q(core_rel__in=self.instance.corerelationship_set.filter(tool__isnull=False))
            self.fields["core_facility_tools"].queryset = Tool.objects.filter(tool_filter | no_facility_filter).exclude(
                parent_tool__isnull=False
            )
            if self.instance.pk:
                self.fields["core_facility_tools"].initial = Tool.objects.filter(tool_filter)
        if "core_facility_areas" in self.fields:
            area_filter = Q()
            if self.instance.pk:
                area_filter = Q(core_rel__in=self.instance.corerelationship_set.filter(area__isnull=False))
            self.fields["core_facility_areas"].queryset = Area.objects.filter(area_filter | no_facility_filter)
            if self.instance.pk:
                self.fields["core_facility_areas"].initial = Area.objects.filter(area_filter)
        if "core_facility_consumables" in self.fields:
            consumable_filter = Q()
            if self.instance.pk:
                consumable_filter = Q(core_rel__in=self.instance.corerelationship_set.filter(consumable__isnull=False))
            self.fields["core_facility_consumables"].queryset = Consumable.objects.filter(
                consumable_filter | no_facility_filter
            )
            if self.instance.pk:
                self.fields["core_facility_consumables"].initial = Consumable.objects.filter(consumable_filter)


@register(CoreFacility)
class CoreFacilityAdmin(admin.ModelAdmin):
    list_display = ["name", "get_external_id"]
    form = CoreFacilityAdminForm

    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)
        if "core_facility_tools" in form.changed_data:
            original_items = Tool.objects.filter(core_rel__in=obj.corerelationship_set.filter(tool__isnull=False))
            save_all_core_facility_relationships(
                form.cleaned_data["core_facility_tools"], original_items, obj, "tool", change
            )
        if "core_facility_areas" in form.changed_data:
            original_items = Area.objects.filter(core_rel__in=obj.corerelationship_set.filter(area__isnull=False))
            save_all_core_facility_relationships(
                form.cleaned_data["core_facility_areas"], original_items, obj, "area", change
            )
        if "core_facility_consumables" in form.changed_data:
            original_items = Consumable.objects.filter(
                core_rel__in=obj.corerelationship_set.filter(consumable__isnull=False)
            )
            save_all_core_facility_relationships(
                form.cleaned_data["core_facility_consumables"], original_items, obj, "consumable", change
            )

    @admin.display(ordering="external_id")
    def get_external_id(self, core_facility: CoreFacility):
        return core_facility.external_id


class NewToolAdminForm(ToolAdminForm):
    core_facility = forms.ModelChoiceField(
        queryset=CoreFacility.objects.all(),
        required=False,
        help_text="The core facility this tool belongs to. Used for billing purposes.",
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.instance.pk and "core_facility" in self.fields:
            self.fields["core_facility"].initial = self.instance.core_facility

    def clean_core_facility(self):
        parent_tool = self.cleaned_data.get("parent_tool")
        core_facility = self.cleaned_data.get("core_facility")
        if not parent_tool and not core_facility and settings.TOOL_CORE_FACILITY_REQUIRED:
            raise ValidationError("This field is required.")
        return core_facility


class NewToolAdmin(ToolAdmin):
    form = NewToolAdminForm

    def save_model(self, request, obj: Tool, form, change):
        super().save_model(request, obj, form, change)
        save_or_delete_core_facility(obj, form.cleaned_data.get("core_facility"), "tool")

    def get_fieldsets(self, request, obj: Area = None):
        # Add core_facility field
        fieldsets = deepcopy(super().get_fieldsets(request, obj))
        fieldsets[0][1]["fields"] = fieldsets[0][1]["fields"] + ("core_facility",)
        return fieldsets


class NewAreaAdminForm(AreaAdminForm):
    core_facility = forms.ModelChoiceField(
        queryset=CoreFacility.objects.all(),
        required=settings.AREA_CORE_FACILITY_REQUIRED,
        help_text="The core facility this area belongs to. Used for billing purposes.",
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.instance.pk and "core_facility" in self.fields:
            self.fields["core_facility"].initial = self.instance.core_facility


class NewAreaAdmin(AreaAdmin):
    form = NewAreaAdminForm

    def get_fieldsets(self, request, obj: Area = None):
        # Add core_facility field
        fieldsets = deepcopy(super().get_fieldsets(request, obj))
        fieldsets[0][1]["fields"] = fieldsets[0][1]["fields"] + ("core_facility",)
        return fieldsets

    def save_model(self, request, obj: Area, form, change):
        super().save_model(request, obj, form, change)
        save_or_delete_core_facility(obj, form.cleaned_data["core_facility"], "area")


class NewConsumableAdminForm(forms.ModelForm):
    core_facility = forms.ModelChoiceField(
        queryset=CoreFacility.objects.all(),
        required=settings.CONSUMABLE_CORE_FACILITY_REQUIRED,
        help_text="The core facility this consumable belongs to. Used for billing purposes.",
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.instance.pk and "core_facility" in self.fields:
            self.fields["core_facility"].initial = self.instance.core_facility


class NewConsumableAdmin(ConsumableAdmin):
    form = NewConsumableAdminForm

    def save_model(self, request, obj: Consumable, form, change):
        super().save_model(request, obj, form, change)
        save_or_delete_core_facility(obj, form.cleaned_data["core_facility"], "consumable")


class NewStaffChargeAdminForm(forms.ModelForm):
    core_facility = forms.ModelChoiceField(
        queryset=CoreFacility.objects.all(),
        required=settings.STAFF_CHARGE_CORE_FACILITY_REQUIRED,
        help_text="The core facility this staff charge belongs to. Used for billing purposes.",
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.instance.pk and "core_facility" in self.fields:
            self.fields["core_facility"].initial = self.instance.core_facility


class NewStaffChargeAdmin(StaffChargeAdmin):
    form = NewStaffChargeAdminForm

    def save_model(self, request, obj: Consumable, form, change):
        super().save_model(request, obj, form, change)
        save_or_delete_core_facility(obj, form.cleaned_data["core_facility"], "staff_charge")


class CustomChargeAdminForm(forms.ModelForm):
    class Meta:
        model = CustomCharge
        fields = "__all__"

    core_facility = forms.ModelChoiceField(
        queryset=CoreFacility.objects.all(),
        required=settings.CUSTOM_CHARGE_CORE_FACILITY_REQUIRED,
        help_text="The core facility this tool belongs to. Used for billing purposes.",
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not cap_discount_installed():
            hide_form_field(self, "cap_eligible")


@register(CustomCharge)
class CustomChargeAdmin(admin.ModelAdmin):
    list_display = ("name", "date", "amount", "customer", "project", "creator", "core_facility", "waived")
    search_fields = ("name", "customer__first_name", "customer__last_name", "customer__username", "project__name")
    list_filter = ("date", "project", "core_facility", "waived")
    form = CustomChargeAdminForm


class ProjectBillingHardCapAdminForm(forms.ModelForm):
    charge_types = IntMultipleChoiceField(
        choices=BillableItemType.choices(),
        required=True,
        widget=widgets.FilteredSelectMultiple(verbose_name="Types", is_stacked=False),
    )

    class Meta:
        model = ProjectBillingHardCap
        fields = "__all__"


@admin.register(ProjectBillingHardCap)
class ProjectBillingHardCapAdmin(admin.ModelAdmin):
    list_display = ["project", "enabled", "start_date", "end_date", "amount", "get_charge_types"]
    list_filter = ["enabled", "project"]
    form = ProjectBillingHardCapAdminForm

    @admin.display(description="Charge types")
    def get_charge_types(self, instance: ProjectBillingHardCap):
        return instance.get_charge_types_display()


# Re-register ToolAdmin, AreaAdmin & ConsumableAdmin
admin.site.unregister(Tool)
admin.site.register(Tool, NewToolAdmin)
admin.site.unregister(Area)
admin.site.register(Area, NewAreaAdmin)
admin.site.unregister(Consumable)
admin.site.register(Consumable, NewConsumableAdmin)
admin.site.unregister(StaffCharge)
admin.site.register(StaffCharge, NewStaffChargeAdmin)
admin.site.register(InstitutionType)
admin.site.register(Department)



import datetime
from datetime import timedelta
from http import HTTPStatus

from NEMO.models import AreaAccessRecord, ConsumableWithdraw, Reservation, StaffCharge, TrainingSession, UsageEvent
from NEMO.policy import policy_class as policy
from NEMO.utilities import get_month_timeframe
from django.core.exceptions import ValidationError
from django.test import TestCase
from django.utils import timezone

from NEMO_billing.exceptions import HardCAPReachedException
from NEMO_billing.invoices.models import BillableItemType, InvoiceConfiguration
from NEMO_billing.models import CustomCharge, ProjectBillingHardCap
from NEMO_billing.tests.cap.test_cap_utilities import create_items
from NEMO_billing.tests.test_utilities import basic_data
from NEMO_billing.utilities import get_charges_amount_between


class TestProjectPrepayments(TestCase):
    def setUp(self) -> None:
        self.user, self.project, self.tool, self.area = basic_data(set_rate_category=True)
        self.project.projectbillingdetails.no_tax = True
        self.project.projectbillingdetails.save()
        self.project_hard_cap = ProjectBillingHardCap(project=self.project)
        self.assertRaises(ValidationError, self.project_hard_cap.full_clean)
        self.project_hard_cap.charge_types = str(BillableItemType.TOOL_USAGE.value)
        self.assertRaises(ValidationError, self.project_hard_cap.full_clean)
        self.project_hard_cap.amount = 30
        self.project_hard_cap.start_date = datetime.date.today()
        self.project_hard_cap.end_date = datetime.date.today()
        self.project_hard_cap.full_clean()
        self.project_hard_cap.save()
        self.tool, self.area, self.consumable, self.staff = create_items()

    def test_non_tax_exempt_configuration(self):
        self.project.projectbillingdetails.no_tax = False
        self.project.projectbillingdetails.save()
        self.assertRaises(ValidationError, self.project_hard_cap.full_clean)
        invoiceConfiguration = InvoiceConfiguration()
        invoiceConfiguration.name = "Config"
        invoiceConfiguration.email_from = "billing@test.com"
        invoiceConfiguration.merchant_name = "Atlantis Labs, LLC"
        invoiceConfiguration.tax = 10
        invoiceConfiguration.save()
        self.project_hard_cap.configuration = invoiceConfiguration
        self.project_hard_cap.full_clean()
        self.project_hard_cap.save()
        # the tool rate is 30 per hour, let's create an hour usage event
        start = timezone.now() - datetime.timedelta(hours=1)
        end = start + datetime.timedelta(hours=1)
        usage = UsageEvent.objects.create(
            user=self.user, operator=self.user, project=self.project, tool=self.tool, start=start, end=end
        )
        self.user.training_required = False
        self.user.save()
        self.user.qualifications.add(self.tool)
        # with taxes, it shouldn't be enough
        response = policy.check_to_enable_tool(
            self.tool, self.user, self.user, project=self.project, staff_charge=False
        )
        self.assertEqual(response.status_code, HTTPStatus.BAD_REQUEST)
        self.assertRaises(
            HardCAPReachedException, policy.check_billing_to_project, self.project, self.user, self.tool, usage
        )
        # 30 tool usage + 10% tax = 33
        start, end = get_month_timeframe(datetime.datetime.now().isoformat())
        self.assertEqual(get_charges_amount_between(self.project, invoiceConfiguration, start, end)[1], 33)
        # test with other dates, yesterday and tomorrow
        self.project_hard_cap.start_date = timezone.now() - timedelta(days=1)
        self.project_hard_cap.end_date = timezone.now() - timedelta(days=1)
        self.project_hard_cap.save()
        response = policy.check_to_enable_tool(
            self.tool, self.user, self.user, project=self.project, staff_charge=False
        )
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.project_hard_cap.start_date = timezone.now() + timedelta(days=1)
        self.project_hard_cap.end_date = timezone.now() + timedelta(days=1)
        self.project_hard_cap.save()
        response = policy.check_to_enable_tool(
            self.tool, self.user, self.user, project=self.project, staff_charge=False
        )
        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_charge_types(self):
        start_charge = timezone.now() - datetime.timedelta(hours=1)
        end_charge = start_charge + datetime.timedelta(hours=1)
        UsageEvent.objects.create(
            user=self.user, operator=self.user, project=self.project, tool=self.tool, start=start_charge, end=end_charge
        )
        start, end = get_month_timeframe(datetime.datetime.now().isoformat())
        self.assertEqual(get_charges_amount_between(self.project, None, start, end)[1], 30)
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.TOOL_USAGE])[1],
            30,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.AREA_ACCESS])[1],
            0,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.STAFF_CHARGE])[1],
            0,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.CUSTOM_CHARGE])[1],
            0,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.CONSUMABLE])[1],
            0,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.TRAINING])[1],
            0,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.MISSED_RESERVATION])[1],
            0,
        )
        # Add area access record. Rate 40
        AreaAccessRecord.objects.create(
            area=self.area, customer=self.user, project=self.project, start=start_charge, end=end_charge
        )
        self.assertEqual(get_charges_amount_between(self.project, None, start, end)[1], 30 + 40)
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.TOOL_USAGE])[1],
            30,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.AREA_ACCESS])[1],
            40,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.STAFF_CHARGE])[1],
            0,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.CUSTOM_CHARGE])[1],
            0,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.CONSUMABLE])[1],
            0,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.TRAINING])[1],
            0,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.MISSED_RESERVATION])[1],
            0,
        )
        # Add ConsumableWithdraw 20 per item
        ConsumableWithdraw.objects.create(
            customer=self.user,
            project=self.project,
            consumable=self.consumable,
            quantity=1,
            date=end,
            merchant=self.staff,
        )
        self.assertEqual(get_charges_amount_between(self.project, None, start, end)[1], 30 + 40 + 20)
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.TOOL_USAGE])[1],
            30,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.AREA_ACCESS])[1],
            40,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.STAFF_CHARGE])[1],
            0,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.CUSTOM_CHARGE])[1],
            0,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.CONSUMABLE])[1],
            20,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.TRAINING])[1],
            0,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.MISSED_RESERVATION])[1],
            0,
        )
        # Add Staff Charge 1h -> 25/hr
        StaffCharge.objects.create(
            staff_member=self.staff,
            customer=self.user,
            project=self.project,
            start=start_charge,
            end=end_charge,
        )
        self.assertEqual(get_charges_amount_between(self.project, None, start, end)[1], 30 + 40 + 20 + 25)
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.TOOL_USAGE])[1],
            30,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.AREA_ACCESS])[1],
            40,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.STAFF_CHARGE])[1],
            25,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.CUSTOM_CHARGE])[1],
            0,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.CONSUMABLE])[1],
            20,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.TRAINING])[1],
            0,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.MISSED_RESERVATION])[1],
            0,
        )
        # Add Custom Charge 5
        CustomCharge.objects.create(customer=self.user, project=self.project, date=end, amount=5, creator=self.staff)
        self.assertEqual(get_charges_amount_between(self.project, None, start, end)[1], 30 + 40 + 20 + 25 + 5)
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.TOOL_USAGE])[1],
            30,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.AREA_ACCESS])[1],
            40,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.STAFF_CHARGE])[1],
            25,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.CUSTOM_CHARGE])[1],
            5,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.CONSUMABLE])[1],
            20,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.TRAINING])[1],
            0,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.MISSED_RESERVATION])[1],
            0,
        )
        # Add training -> rate 10
        TrainingSession.objects.create(
            trainee=self.user,
            project=self.project,
            date=end,
            trainer=self.staff,
            duration=60,
            tool=self.tool,
            type=TrainingSession.Type.INDIVIDUAL,
        )
        self.assertEqual(get_charges_amount_between(self.project, None, start, end)[1], 30 + 40 + 20 + 25 + 5 + 10)
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.TOOL_USAGE])[1],
            30,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.AREA_ACCESS])[1],
            40,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.STAFF_CHARGE])[1],
            25,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.CUSTOM_CHARGE])[1],
            5,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.CONSUMABLE])[1],
            20,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.TRAINING])[1],
            10,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.MISSED_RESERVATION])[1],
            0,
        )
        # Add missed reservation -> rate 15
        Reservation.objects.create(
            user=self.user,
            creator=self.user,
            project=self.project,
            start=start_charge,
            end=end_charge,
            missed=True,
            tool=self.tool,
            short_notice=False,
        )
        self.assertEqual(get_charges_amount_between(self.project, None, start, end)[1], 30 + 40 + 20 + 25 + 5 + 10 + 15)
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.TOOL_USAGE])[1],
            30,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.AREA_ACCESS])[1],
            40,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.STAFF_CHARGE])[1],
            25,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.CUSTOM_CHARGE])[1],
            5,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.CONSUMABLE])[1],
            20,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.TRAINING])[1],
            10,
        )
        self.assertEqual(
            get_charges_amount_between(self.project, None, start, end, [BillableItemType.MISSED_RESERVATION])[1],
            15,
        )



from NEMO.mixins import ModelAdminRedirectMixin
from NEMO.utilities import new_model_copy
from django.contrib import admin
from django.contrib.admin import register
from django.utils.safestring import mark_safe

from NEMO_billing.quotes.models import Quote, QuoteConfiguration, QuoteItem
from NEMO_billing.quotes.utilities import delete_approval_request_notification


def clone_configuration(configuration: QuoteConfiguration):
    """
    Create a clone of selected configurations and save them.

    :param configuration:
    :return: Newly cloned configuration
    """
    new_configuration = new_model_copy(configuration)
    new_configuration.name = f"Copy of {configuration.name}"
    new_configuration.save()

    return new_configuration


@admin.action(description="Clone selected Configurations")
def clone_configurations(modeladmin, request, queryset):
    for configuration in queryset.all():
        clone_configuration(configuration)


@register(QuoteConfiguration)
class QuoteConfigurationAdmin(ModelAdminRedirectMixin, admin.ModelAdmin):
    list_display = ("id", "name", "merchant_name", "currency", "get_create_permission", "get_approval_permissions")
    ordering = ("name",)
    actions = (clone_configurations,)
    fieldsets = (
        (None, {"fields": ("name", "expiration_in_days")}),
        (
            "Quote Number template",
            {"fields": ("quote_numbering_template", "current_quote_number")},
        ),
        ("Communication", {"fields": ("email_cc",)}),
        ("Permissions", {"fields": ("create_permissions", "approval_permissions")}),
        ("Merchant info", {"fields": ("merchant_name", "merchant_details", "merchant_logo", "terms")}),
        ("Currency", {"fields": ("currency", "currency_symbol")}),
        ("Tax", {"fields": ("tax_name", "tax")}),
    )

    @admin.display(description="Create permissions", ordering="create_permissions")
    def get_create_permission(self, obj: QuoteConfiguration):
        return mark_safe(
            "<br>".join(
                str(obj.get_create_permissions_field().role_display(role_str, admin_display=True))
                for role_str in obj.create_permissions
            )
        )

    @admin.display(description="Approval permissions", ordering="approval_permissions")
    def get_approval_permissions(self, obj: QuoteConfiguration):
        return mark_safe(
            "<br>".join(
                str(obj.get_approval_permissions_field().role_display(role_str, admin_display=True))
                for role_str in obj.approval_permissions
            )
        )


@admin.action(description="Unpublish selected quotes")
def unpublish_quote(modeladmin, request, queryset):
    quotes = queryset.filter(status=Quote.Status.PUBLISHED)
    quotes.update(
        status=Quote.Status.DRAFT,
        approved_by=None,
        expiration_date=None,
        published_date=None,
        file=None,
        file_access_token=None,
    )
    for quote in quotes:
        delete_approval_request_notification(quote)


class QuoteItemInline(admin.TabularInline):
    model = QuoteItem
    can_delete = False

    def has_change_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request, obj=None):
        return False


@register(Quote)
class QuoteAdmin(admin.ModelAdmin):
    inlines = [QuoteItemInline]
    list_display = ("id", "name", "quote_number", "status", "configuration")
    actions = [unpublish_quote]
    ordering = ("name",)
    list_filter = ("configuration", "status")
    readonly_fields = ["creator", "created_date", "updated_date", "file", "file_access_token"]

    def delete_queryset(self, request, queryset):
        # forcing it to go through the model delete so the notifications are deleted as well
        for quote in queryset:
            quote.delete()



from datetime import datetime
from typing import Dict

from NEMO.decorators import customization
from NEMO.utilities import get_month_timeframe
from NEMO.views.customization import CustomizationBase
from django.core.exceptions import ValidationError
from django.template import Context, Template


@customization(key="invoices", title="Billing invoices")
class InvoiceCustomization(CustomizationBase):
    variables = {
        "invoice_number_format": "{:05d}",
        "invoice_number_current": "0",
        "invoice_zip_export_template": "{{ account_name }}/{{ project_name }}-{{ invoice.start|date:'Y' }}-{{ invoice.start|date:'F' }}-{{ invoice.invoice_number|slugify }}",
        "invoice_month_list_since": "",
        "invoice_funds_show_total_balance": "",
        "invoice_show_hard_cap_status": "",
        "invoice_skip_inactive_projects": "",
        "invoice_skip_inactive_accounts": "",
        "invoice_send_email_when_voided": "",
    }
    files = [
        ("email_send_invoice_subject", ".txt"),
        ("email_send_invoice_message", ".html"),
        ("email_send_invoice_voided_subject", ".txt"),
        ("email_send_invoice_voided_message", ".html"),
        ("email_send_invoice_reminder_subject", ".txt"),
        ("email_send_invoice_reminder_message", ".html"),
    ]

    def context(self) -> Dict:
        # Adding invoice number formatted to the template
        customization_context = super().context()
        try:
            customization_context["invoice_number_formatted"] = self.get("invoice_number_format").format(
                int(self.get("invoice_number_current"))
            )
        except:
            pass
        try:
            customization_context["invoice_zip_template_formatted"] = (
                self.invoice_template_validation_context(self.get("invoice_zip_export_template")) + ".pdf"
            )
        except:
            pass
        return customization_context

    def validate(self, name, value):
        if name == "invoice_number_format":
            try:
                value.format(123)
            except Exception as e:
                raise ValidationError(str(e))
        elif name == "invoice_zip_export_template":
            try:
                self.invoice_template_validation_context(value)
            except Exception as e:
                raise ValidationError(str(e))
        elif name == "invoice_month_list_since" and value:
            self.validate_date(value)

    def invoice_template_validation_context(self, value):
        from NEMO_billing.invoices.models import Invoice

        invoice = Invoice()
        invoice.invoice_number = invoice.generate_invoice_number()
        month = datetime.now().month - 1
        if month == 0:
            month = 12
        invoice_date = datetime.now().replace(month=month)
        invoice.start, invoice.end = get_month_timeframe(invoice_date.isoformat())
        invoice.created_date = datetime.now()
        return Template(value).render(
            Context({"invoice": invoice, "project_name": "test-project", "account_name": "test-account"})
        )



# Generated by Django 3.2.16 on 2023-01-13 23:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("NEMO_billing", "0002_version_1_2_0"),
    ]

    operations = [
        migrations.AddField(
            model_name="customcharge",
            name="cap_eligible",
            field=models.BooleanField(default=False, help_text="Check this box to make this charge count towards CAP"),
        ),
    ]



# Generated by Django 2.2.13 on 2021-03-01 16:49

import django.db.models.deletion
import mptt.fields
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("NEMO", "0027_version_3_8_0"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    def create_rate_types(apps, schema_editor):
        RateTypeModel = apps.get_model("rates", "RateType")
        if not RateTypeModel.objects.exists():
            RateTypeModel.objects.create(type="TOOL_USAGE", category_specific=True, item_specific=True)
            RateTypeModel.objects.create(type="TOOL_TRAINING_INDIVIDUAL", category_specific=True, item_specific=True)
            RateTypeModel.objects.create(type="TOOL_TRAINING_GROUP", category_specific=True, item_specific=True)
            RateTypeModel.objects.create(type="TOOL_MISSED_RESERVATION", category_specific=False, item_specific=False)
            RateTypeModel.objects.create(type="AREA_USAGE", category_specific=True, item_specific=True)
            RateTypeModel.objects.create(type="AREA_MISSED_RESERVATION", category_specific=False, item_specific=False)
            RateTypeModel.objects.create(type="CONSUMABLE", category_specific=False, item_specific=True)
            RateTypeModel.objects.create(type="STAFF_CHARGE", category_specific=True, item_specific=False)

    operations = [
        migrations.CreateModel(
            name="RateCategory",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(help_text="The name of this rate category", max_length=200)),
            ],
            options={
                "verbose_name": "Rate category",
                "verbose_name_plural": "Rate categories",
            },
        ),
        migrations.CreateModel(
            name="RateType",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "type",
                    models.CharField(
                        choices=[
                            (
                                "Tool",
                                (
                                    ("TOOL_USAGE", "Tool usage"),
                                    ("TOOL_TRAINING_INDIVIDUAL", "Tool individual training"),
                                    ("TOOL_TRAINING_GROUP", "Tool group training"),
                                    ("TOOL_MISSED_RESERVATION", "Tool missed reservation"),
                                ),
                            ),
                            (
                                "Area",
                                (("AREA_USAGE", "Area usage"), ("AREA_MISSED_RESERVATION", "Area missed reservation")),
                            ),
                            ("CONSUMABLE", "Consumable/Supply"),
                            ("STAFF_CHARGE", "Staff charge"),
                        ],
                        max_length=100,
                    ),
                ),
                (
                    "category_specific",
                    models.BooleanField(
                        default=False,
                        help_text="Check this box to make this rate type category specific (i.e. you will need to enter a rate for each category)",
                    ),
                ),
                (
                    "item_specific",
                    models.BooleanField(
                        default=False,
                        help_text="Check this box to make this rate type item specific (i.e. you will need to enter a rate for each item)",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="RateLog",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("object_id", models.PositiveIntegerField(null=True)),
                ("action", models.IntegerField(choices=[(0, "Add"), (1, "Delete"), (2, "Update")])),
                ("date", models.DateTimeField(auto_now_add=True)),
                ("details", models.TextField(blank=True, null=True)),
                (
                    "content_type",
                    models.ForeignKey(
                        null=True, on_delete=django.db.models.deletion.SET_NULL, to="contenttypes.ContentType"
                    ),
                ),
                ("user", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name="Rate",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "amount",
                    models.DecimalField(
                        decimal_places=2, max_digits=8, help_text="The rate per hour if this isn't a flat rate"
                    ),
                ),
                (
                    "flat",
                    models.BooleanField(
                        default=False, help_text="Check this box to make this a flat rate (independent of time)"
                    ),
                ),
                (
                    "minimum_charge",
                    models.DecimalField(
                        decimal_places=2,
                        max_digits=8,
                        blank=True,
                        help_text="The minimum charge for this rate",
                        null=True,
                    ),
                ),
                (
                    "area",
                    mptt.fields.TreeForeignKey(
                        blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to="NEMO.Area"
                    ),
                ),
                (
                    "category",
                    models.ForeignKey(
                        blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to="rates.RateCategory"
                    ),
                ),
                (
                    "consumable",
                    models.ForeignKey(
                        blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to="NEMO.Consumable"
                    ),
                ),
                (
                    "tool",
                    models.ForeignKey(
                        blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to="NEMO.Tool"
                    ),
                ),
                ("type", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="rates.RateType")),
            ],
        ),
        migrations.RunPython(create_rate_types),
    ]



import importlib
from abc import ABC, abstractmethod
from decimal import Decimal
from io import BytesIO, IOBase
from typing import Any, List, Tuple

from NEMO.utilities import BasicDisplayTable
from NEMO.views.customization import ApplicationCustomization, ProjectsAccountsCustomization
from django.conf import settings
from django.db.models import Sum
from django.utils import timezone

from NEMO_billing.invoices.exceptions import (
    InvoiceGenerationException,
)
from NEMO_billing.invoices.models import (
    BillableItemType,
    Invoice,
    InvoiceDetailItem,
    InvoiceSummaryItem,
)
from NEMO_billing.invoices.utilities import category_name_for_item_type, display_amount
from NEMO_billing.pdf_utilities import (
    HEADING_FONT,
    HEADING_FONT_CENTERED,
    PDFDocument,
    PDFPageSize,
    SMALL_FONT,
    SMALL_FONT_RIGHT,
)


class InvoiceRenderer(ABC):
    def __init__(self):
        self.date_time_format = settings.INVOICE_DATETIME_FORMAT
        self.date_format = settings.INVOICE_DATE_FORMAT

    def render_invoice(self, invoice: Invoice) -> IOBase:
        try:
            file_bytes = BytesIO()
            document = self.init_document(invoice, file_bytes)
            self.render_front_page(invoice, document)
            self.render_summary(invoice, document)
            if invoice.configuration.detailed_invoice:
                self.render_details(invoice, document)
            return self.close_document(invoice, document, file_bytes)
        except Exception as e:
            raise InvoiceGenerationException(invoice, self, e)

    @abstractmethod
    def get_file_extension(self):
        pass

    @abstractmethod
    def init_document(self, invoice: Invoice, file_bytes: IOBase) -> Any:
        pass

    @abstractmethod
    def render_front_page(self, invoice: Invoice, document):
        pass

    @abstractmethod
    def render_summary(self, invoice: Invoice, document):
        pass

    @abstractmethod
    def render_details(self, invoice: Invoice, document):
        pass

    def close_document(self, invoice: Invoice, document, file_bytes: IOBase) -> IOBase:
        return file_bytes

    def format_date_time(self, datetime, date_only: bool = False):
        tz_datetime = datetime.astimezone(timezone.get_current_timezone())
        return tz_datetime.strftime(self.date_format) if date_only else tz_datetime.strftime(self.date_time_format)


class PDFInvoiceRenderer(InvoiceRenderer):
    def get_file_extension(self):
        return "pdf"

    def init_document(self, invoice: Invoice, file_bytes) -> PDFDocument:
        pdf = PDFDocument(
            buffer_bytes=file_bytes,
            pagesize=PDFPageSize.Letter,
            header_text=invoice.configuration.merchant_details,
            header_logo=invoice.configuration.merchant_logo,
            document_author=invoice.created_by.get_name(),
            document_title=f"{invoice.configuration.invoice_title} {invoice.invoice_number}",
        )
        return pdf

    def render_front_page(self, invoice: Invoice, pdf: PDFDocument):
        default_style = pdf.styles[f"{SMALL_FONT}Base11"]
        default_italic_style = pdf.styles[f"{SMALL_FONT}Italic10"]
        heading_style = pdf.styles[f"{HEADING_FONT}Bold18"]
        heading_style_centered = pdf.styles[f"{HEADING_FONT_CENTERED}Bold18"]
        details_key_style = pdf.styles[f"{SMALL_FONT}Bold12"]
        total_charges_style = pdf.styles[f"{SMALL_FONT}Base12"]
        pdf.add_space(1, 10)
        pdf.add_title(invoice.configuration.invoice_title)
        key_value_col_width = [65, None]
        pdf.add_space(1, 30)
        details_data = [
            ("Date:", self.format_date_time(invoice.created_date, date_only=True)),
            ("Number:", invoice.invoice_number),
            ("Project:", invoice.project_details.name),
            (
                ProjectsAccountsCustomization.get("project_application_identifier_name") + ":",
                invoice.project_details.project.application_identifier,
            ),
        ]
        pdf.add_table_key_value(details_data, key_value_col_width, details_key_style, default_style)
        pdf.add_space(1, 15)
        if invoice.project_details.addressee:
            pdf.add_table_key_value(
                [("To:", invoice.project_details.addressee)], key_value_col_width, details_key_style, default_style
            )
        pdf.add_space(1, 50)
        facility = ApplicationCustomization.get("facility_name")
        pdf.add_heading_paragraph(
            f"{facility} charges for {invoice.start.strftime(self.date_format)} to {invoice.end.strftime(self.date_format)}",
            heading_style_centered,
            fit=True,
            border=True,
        )
        pdf.add_space(1, 15)
        pdf.add_paragraph(
            f"Total {facility} charges: <b>{invoice.total_amount_display()}</b>", style=total_charges_style
        )
        pdf.add_space(1, 5)
        pdf.add_paragraph(f"Please refer to attached summary for details.", default_italic_style)
        if invoice.configuration.terms:
            pdf.add_space(1, 70)
            pdf.add_heading_paragraph("Terms and Conditions:", heading_style)
            pdf.add_paragraph(invoice.configuration.terms, default_style)

    def render_summary(self, invoice: Invoice, pdf: PDFDocument):
        header_name_style = pdf.styles[f"{SMALL_FONT}Bold10"]
        header_amount_style = pdf.styles[f"{SMALL_FONT_RIGHT}Bold10"]
        category_name_style = pdf.styles[f"{SMALL_FONT}Bold12"]
        category_amount_style = pdf.styles[f"{SMALL_FONT_RIGHT}Base12"]
        item_amount_style = pdf.styles[f"{SMALL_FONT_RIGHT}Base10"]
        other_name_style = pdf.styles[f"{SMALL_FONT}Bold10"]
        total_amount_style = pdf.styles[f"{SMALL_FONT_RIGHT}Bold12"]
        pdf.add_page_break()
        pdf.add_title("Summary")
        table_data = []
        # Separate type based on invoice configuration options
        separate_item_types: List[BillableItemType] = []
        facility_item_types: List[BillableItemType] = []

        if invoice.configuration.separate_tool_usage_charges:
            separate_item_types.append(BillableItemType.TOOL_USAGE)
        else:
            facility_item_types.append(BillableItemType.TOOL_USAGE)
        if invoice.configuration.separate_area_access_charges:
            separate_item_types.append(BillableItemType.AREA_ACCESS)
        else:
            facility_item_types.append(BillableItemType.AREA_ACCESS)
        if invoice.configuration.separate_consumable_charges:
            separate_item_types.append(BillableItemType.CONSUMABLE)
        else:
            facility_item_types.append(BillableItemType.CONSUMABLE)
        if invoice.configuration.separate_missed_reservation_charges:
            separate_item_types.append(BillableItemType.MISSED_RESERVATION)
        else:
            facility_item_types.append(BillableItemType.MISSED_RESERVATION)
        if invoice.configuration.separate_staff_charges:
            separate_item_types.append(BillableItemType.STAFF_CHARGE)
        else:
            facility_item_types.append(BillableItemType.STAFF_CHARGE)
        if invoice.configuration.separate_training_charges:
            separate_item_types.append(BillableItemType.TRAINING)
        else:
            facility_item_types.append(BillableItemType.TRAINING)
        if invoice.configuration.separate_custom_charges:
            separate_item_types.append(BillableItemType.CUSTOM_CHARGE)
        else:
            facility_item_types.append(BillableItemType.CUSTOM_CHARGE)

        # Deal first with non-empty core facilities
        for core_facility in invoice.sorted_core_facilities():
            if core_facility:
                facility_summaries = invoice.invoicesummaryitem_set.filter(core_facility=core_facility)
                summary_items = facility_summaries.filter(
                    summary_item_type=InvoiceSummaryItem.InvoiceSummaryItemType.ITEM
                )
                facility_display_items = summary_items.filter(item_type__in=[t.value for t in facility_item_types])
                sub_total_amount = facility_display_items.aggregate(Sum("amount"))["amount__sum"]
                # Don't show the facility if there are no items in it
                if facility_display_items.exists():
                    table_data.append(["", "", ""])
                    table_data.append(
                        [(core_facility, category_name_style), ("", category_name_style), ("", category_amount_style)]
                    )
                    for item_type in facility_item_types:
                        items = summary_items.filter(item_type=item_type.value)
                        self.add_items(invoice.configuration, table_data, items, pdf)
                    discounts_amount = self.add_discounts(invoice.configuration, table_data, facility_summaries, pdf)
                    table_data.append(
                        [
                            ("Subtotal", other_name_style),
                            "",
                            (
                                display_amount(sub_total_amount + discounts_amount, invoice.configuration),
                                item_amount_style,
                            ),
                        ]
                    )

        # Add items not in any facilities (and items set as separate in configuration)
        general_summaries = invoice.invoicesummaryitem_set.filter(core_facility=None)
        general_summary_items = general_summaries.filter(
            summary_item_type=InvoiceSummaryItem.InvoiceSummaryItemType.ITEM
        )
        for item_type in facility_item_types + separate_item_types:
            if item_type in facility_item_types:
                items = general_summary_items.filter(item_type=item_type.value)
            else:
                items = invoice.invoicesummaryitem_set.filter(item_type=item_type.value)
            if items.exists():
                table_data.append(["", "", ""])
                table_data.append(
                    [(category_name_for_item_type(item_type), category_name_style), "", ("", category_amount_style)]
                )
                self.add_items(invoice.configuration, table_data, items, pdf)
                sub_total_amount = items.aggregate(Sum("amount"))["amount__sum"]
                discounts_amount = self.add_discounts(invoice.configuration, table_data, general_summaries, pdf)
                table_data.append(
                    [
                        ("Subtotal", other_name_style),
                        "",
                        (display_amount(sub_total_amount + discounts_amount, invoice.configuration), item_amount_style),
                    ]
                )

        total_charges_amount = invoice.invoicesummaryitem_set.filter(
            summary_item_type=InvoiceSummaryItem.InvoiceSummaryItemType.SUBTOTAL
        ).aggregate(Sum("amount"))["amount__sum"]
        table_data.append(["", "", ""])
        table_data.append(
            [
                ("Total Charges", category_name_style),
                "",
                (display_amount(total_charges_amount, invoice.configuration), category_amount_style),
            ]
        )

        tax_item = invoice.invoicesummaryitem_set.filter(
            summary_item_type=InvoiceSummaryItem.InvoiceSummaryItemType.TAX
        ).first()
        if tax_item:
            table_data.append(
                [
                    (tax_item.name, category_name_style),
                    tax_item.details or "",
                    (display_amount(tax_item.amount, invoice.configuration), category_amount_style),
                ]
            )

        fund_items = invoice.invoicesummaryitem_set.filter(
            summary_item_type=InvoiceSummaryItem.InvoiceSummaryItemType.FUND
        )
        for fund_summary in fund_items:
            table_data.append(
                [
                    (fund_summary.name, category_name_style),
                    "",  # Don't show details for Fund (it contains the id)
                    (display_amount(fund_summary.amount, invoice.configuration), category_amount_style),
                ]
            )

        other_items = invoice.invoicesummaryitem_set.filter(
            summary_item_type=InvoiceSummaryItem.InvoiceSummaryItemType.OTHER
        )
        for other_summary in other_items:
            table_data.append(
                [
                    (other_summary.name, other_name_style),
                    "",  # Don't show details for Fund (it contains the id)
                    (display_amount(other_summary.amount, invoice.configuration), category_amount_style),
                ]
            )

        table_data.append(
            [
                ("Grand Total", category_name_style),
                "",
                (display_amount(invoice.total_amount, invoice.configuration), total_amount_style),
            ]
        )
        summary_col_width = [None, 87, 87]
        pdf.add_table(
            table_data,
            headers=[("Item", header_name_style), ("Rate", header_name_style), ("Amount Due", header_amount_style)],
            col_width=summary_col_width,
            grid=False,
            border=True,
        )

    def add_discounts(self, configuration, table_data, summaries_qs, pdf) -> Decimal:
        discount_name_style = pdf.styles[f"{SMALL_FONT}Bold9"]
        discounts_qs = summaries_qs.filter(summary_item_type=InvoiceSummaryItem.InvoiceSummaryItemType.DISCOUNT)
        item_amount_style = pdf.styles[f"{SMALL_FONT_RIGHT}Base10"]
        total_discounts_amount = Decimal(0)
        for facility_discount in discounts_qs:
            facility_discount_amount = facility_discount.amount
            table_data.append(
                [
                    (facility_discount.name, discount_name_style),
                    (facility_discount.details or ""),
                    (display_amount(facility_discount_amount, configuration), item_amount_style),
                ]
            )
            total_discounts_amount += facility_discount_amount
        return total_discounts_amount

    def add_items(self, configuration, table_data, summary_items_qs, pdf):
        item_amount_style = pdf.styles[f"{SMALL_FONT_RIGHT}Base10"]
        for item in summary_items_qs:
            if not configuration.hide_zero_charge or item.amount != Decimal(0):
                # Add prefix of category name only if we have a facility (otherwise it will be in the category with that name)
                item_name = (
                    f"{category_name_for_item_type(item.item_type)} - {item.name}"
                    if item.core_facility
                    else f"{item.name}"
                )
                table_data.append(
                    [item_name, item.details or "", (display_amount(item.amount, configuration), item_amount_style)]
                )

    def render_details(self, invoice: Invoice, pdf: PDFDocument):
        default_style = pdf.styles[f"{SMALL_FONT}Base12"]
        heading_style_centered = pdf.styles[f"{HEADING_FONT_CENTERED}Bold18"]
        pdf.add_page_break()
        seven_col_width = [None, None, None, None, 65, 77, 50]
        six_col_width = [None, None, None, 65, 77, 50]
        four_col_width = [None, None, None, 50]
        for core_facility in invoice.sorted_core_facilities():
            facility_name = core_facility or "General Charges"
            pdf.add_space(1, 20)
            pdf.add_heading_paragraph(facility_name, heading_style_centered, border=True)
            pdf.add_space(1, 20)
            tool_details = invoice.tool_usage_details(core_facility=core_facility)
            if tool_details.exists():
                pdf.add_space(1, 10)
                pdf.add_paragraph(category_name_for_item_type(BillableItemType.TOOL_USAGE), default_style)
                pdf.add_table(
                    *self.duration_details_data(tool_details, "Tool", pdf), col_width=seven_col_width, grid=False
                )
            area_details = invoice.area_access_details(core_facility=core_facility)
            if area_details.exists():
                pdf.add_space(1, 10)
                pdf.add_paragraph(category_name_for_item_type(BillableItemType.AREA_ACCESS), default_style)
                pdf.add_table(
                    *self.duration_details_data(area_details, "Area", pdf), col_width=seven_col_width, grid=False
                )
            staff_charges_details = invoice.staff_charge_details(core_facility=core_facility)
            if staff_charges_details.exists():
                pdf.add_space(1, 10)
                pdf.add_paragraph(category_name_for_item_type(BillableItemType.STAFF_CHARGE), default_style)
                pdf.add_table(
                    *self.duration_details_data(staff_charges_details, "Item", pdf),
                    col_width=seven_col_width,
                    grid=False,
                )
            consumable_withdrawals = invoice.consumable_withdrawal_details(core_facility=core_facility)
            if consumable_withdrawals.exists():
                pdf.add_space(1, 10)
                pdf.add_paragraph(category_name_for_item_type(BillableItemType.CONSUMABLE), default_style)
                pdf.add_table(
                    *self.consumable_details_data(consumable_withdrawals, pdf), col_width=six_col_width, grid=False
                )
            training_details = invoice.training_details(core_facility=core_facility)
            if training_details.exists():
                pdf.add_space(1, 10)
                pdf.add_paragraph(category_name_for_item_type(BillableItemType.TRAINING), default_style)
                pdf.add_table(*self.training_details_data(training_details, pdf), col_width=six_col_width, grid=False)
            missed_details = invoice.missed_reservation_details(core_facility=core_facility)
            if missed_details.exists():
                pdf.add_space(1, 10)
                pdf.add_paragraph(category_name_for_item_type(BillableItemType.MISSED_RESERVATION), default_style)
                pdf.add_table(
                    *self.duration_details_data(missed_details, "Reservation", pdf),
                    col_width=seven_col_width,
                    grid=False,
                )
            custom_details = invoice.custom_charges_details(core_facility=core_facility)
            if custom_details.exists():
                pdf.add_space(1, 10)
                pdf.add_paragraph(category_name_for_item_type(BillableItemType.CUSTOM_CHARGE), default_style)
                pdf.add_table(
                    *self.custom_charge_details_data(custom_details, pdf), col_width=four_col_width, grid=False
                )

    def close_document(self, invoice: Invoice, pdf: PDFDocument, file_bytes: IOBase) -> IOBase:
        pdf.build()
        return super().close_document(invoice, pdf, file_bytes)

    def duration_details_data(self, items: List[InvoiceDetailItem], item_name: str, pdf) -> Tuple[List[List], List]:
        last_item_style = pdf.styles[f"{SMALL_FONT_RIGHT}Base8"]
        last_header_style = pdf.styles[f"{SMALL_FONT_RIGHT}Bold10"]
        return (
            [
                [
                    str(item.user),
                    item.name,
                    self.format_date_time(item.start),
                    self.format_date_time(item.end),
                    item.quantity_display(),
                    item.rate,
                    (item.amount_display(), last_item_style),
                ]
                for item in items
            ],
            ["User", item_name, "Start Time", "End Time", "Quantity", "Rate", ("Amount", last_header_style)],
        )

    def consumable_details_data(self, items: List[InvoiceDetailItem], pdf) -> Tuple[List[List], List]:
        last_item_style = pdf.styles[f"{SMALL_FONT_RIGHT}Base8"]
        last_header_style = pdf.styles[f"{SMALL_FONT_RIGHT}Bold10"]
        return (
            [
                [
                    str(item.user),
                    item.name,
                    self.format_date_time(item.start),
                    item.quantity_display(),
                    item.rate,
                    (item.amount_display(), last_item_style),
                ]
                for item in items
            ],
            ["User", "Supply", "Date", "Quantity", "Rate", ("Amount", last_header_style)],
        )

    def training_details_data(self, items: List[InvoiceDetailItem], pdf) -> Tuple[List[List], List]:
        last_item_style = pdf.styles[f"{SMALL_FONT_RIGHT}Base8"]
        last_header_style = pdf.styles[f"{SMALL_FONT_RIGHT}Bold10"]
        return (
            [
                [
                    str(item.user),
                    "Tool",
                    self.format_date_time(item.start),
                    item.quantity_display(),
                    item.rate,
                    (item.amount_display(), last_item_style),
                ]
                for item in items
            ],
            ["User", "Tool", "Date", "Duration", "Rate", ("Amount", last_header_style)],
        )

    def custom_charge_details_data(self, items: List[InvoiceDetailItem], pdf) -> Tuple[List[List], List]:
        last_item_style = pdf.styles[f"{SMALL_FONT_RIGHT}Base8"]
        last_header_style = pdf.styles[f"{SMALL_FONT_RIGHT}Bold10"]
        return (
            [
                [str(item.user), item.name, self.format_date_time(item.start), (item.amount_display(), last_item_style)]
                for item in items
            ],
            ["User", "Charge", "Date", ("Amount", last_header_style)],
        )


class CSVInvoiceRenderer(InvoiceRenderer):
    PROJECT_ID_KEY = "application_identifier"
    ACCOUNT_TYPE_KEY = "account_type"
    FACILITY_KEY = "facility"
    USER_KEY = "user"
    ITEM_TYPE_KEY = "item_type"
    ITEM_KEY = "item"
    START_KEY = "start"
    END_KEY = "end"
    QUANTITY_KEY = "quantity"
    RATE_KEY = "rate"
    DISCOUNT_KEY = "discount"
    AMOUNT_KEY = "amount"

    def __init__(self):
        super().__init__()

    def get_file_extension(self):
        return "csv"

    def init_document(self, invoice: Invoice, file_bytes) -> Any:
        document = BasicDisplayTable()
        document.add_header(
            (self.PROJECT_ID_KEY, ProjectsAccountsCustomization.get("project_application_identifier_name"))
        )
        if invoice.project_details.project.account.type:
            document.add_header((self.ACCOUNT_TYPE_KEY, "Account type"))
        if invoice.invoicedetailitem_set.filter(core_facility__isempty=False).exists():
            document.add_header((self.FACILITY_KEY, "Facility"))
        document.add_header((self.USER_KEY, "User"))
        document.add_header((self.ITEM_TYPE_KEY, "Item Type"))
        document.add_header((self.ITEM_KEY, "Item"))
        document.add_header((self.START_KEY, "Start Time"))
        document.add_header((self.END_KEY, "End Time"))
        document.add_header((self.QUANTITY_KEY, "Quantity"))
        document.add_header((self.RATE_KEY, "Rate"))
        document.add_header((self.DISCOUNT_KEY, "Discount"))
        document.add_header((self.AMOUNT_KEY, "Amount"))
        return document

    def render_front_page(self, invoice: Invoice, document):
        pass

    def render_summary(self, invoice: Invoice, document):
        pass

    def render_details(self, invoice: Invoice, document: BasicDisplayTable):
        for item in invoice.invoice_details().order_by("-start"):
            document.add_row(
                {
                    self.PROJECT_ID_KEY: invoice.project_details.project.application_identifier,
                    self.ACCOUNT_TYPE_KEY: invoice.project_details.project.account.type,
                    self.FACILITY_KEY: item.core_facility,
                    self.USER_KEY: item.user,
                    self.ITEM_TYPE_KEY: item.get_item_type_display(),
                    self.ITEM_KEY: item.name,
                    self.START_KEY: self.format_date_time(item.start),
                    self.END_KEY: self.format_date_time(item.end),
                    self.QUANTITY_KEY: item.quantity,
                    self.RATE_KEY: item.rate,
                    self.DISCOUNT_KEY: "waived" if item.waived else item.discount,
                    self.AMOUNT_KEY: item.amount,
                }
            )
        for discount in invoice.invoicesummaryitem_set.filter(
            summary_item_type=InvoiceSummaryItem.InvoiceSummaryItemType.DISCOUNT
        ):
            document.add_row(
                {
                    self.PROJECT_ID_KEY: invoice.project_details.project.application_identifier,
                    self.ACCOUNT_TYPE_KEY: invoice.project_details.project.account.type,
                    self.FACILITY_KEY: discount.core_facility,
                    self.ITEM_TYPE_KEY: discount.get_item_type_display(),
                    self.ITEM_KEY: discount.name,
                    self.AMOUNT_KEY: discount.amount,
                }
            )
        for other in invoice.invoicesummaryitem_set.filter(
            summary_item_type=InvoiceSummaryItem.InvoiceSummaryItemType.OTHER
        ):
            document.add_row(
                {
                    self.PROJECT_ID_KEY: invoice.project_details.project.application_identifier,
                    self.ACCOUNT_TYPE_KEY: invoice.project_details.project.account.type,
                    self.FACILITY_KEY: other.core_facility,
                    self.ITEM_TYPE_KEY: other.get_item_type_display(),
                    self.ITEM_KEY: other.name,
                    self.AMOUNT_KEY: other.amount,
                }
            )
        for tax in invoice.invoicesummaryitem_set.filter(
            summary_item_type=InvoiceSummaryItem.InvoiceSummaryItemType.TAX
        ):
            document.add_row(
                {
                    self.FACILITY_KEY: tax.core_facility,
                    self.ITEM_TYPE_KEY: tax.get_item_type_display(),
                    self.ITEM_KEY: tax.name,
                    self.AMOUNT_KEY: tax.amount,
                }
            )
        for fund in invoice.invoicesummaryitem_set.filter(
            summary_item_type=InvoiceSummaryItem.InvoiceSummaryItemType.FUND
        ):
            document.add_row(
                {
                    self.ITEM_TYPE_KEY: fund.get_item_type_display(),
                    self.ITEM_KEY: fund.name,
                    self.AMOUNT_KEY: fund.amount,
                }
            )
        document.add_row({self.DISCOUNT_KEY: "Total", self.AMOUNT_KEY: invoice.total_amount})

    def close_document(self, invoice: Invoice, document: BasicDisplayTable, file_bytes: IOBase):
        # write the bytes and let the parent class return it
        file_bytes = BytesIO(document.to_csv().content)
        return super().close_document(invoice, document, file_bytes)


def get_invoice_renderer_class() -> InvoiceRenderer:
    renderer_class = getattr(settings, "INVOICE_RENDERER_CLASS", "NEMO_billing.invoices.renderers.PDFInvoiceRenderer")
    assert isinstance(renderer_class, str)
    pkg, attr = renderer_class.rsplit(".", 1)
    ret = getattr(importlib.import_module(pkg), attr)
    return ret()


invoice_renderer_class = get_invoice_renderer_class()



from NEMO.decorators import customization
from NEMO.views.customization import CustomizationBase


@customization(key="quotes", title="Billing quotes")
class QuoteCustomization(CustomizationBase):
    variables = {}
    files = [
        ("quote_approval_request_email", ".html"),
        ("quote_status_update_email", ".html"),
        ("quote_email", ".html"),
    ]



# Generated by Django 4.2.11 on 2024-08-27 15:58

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("NEMO_billing", "0006_projectbillinghardcap"),
    ]

    operations = [
        migrations.AlterField(
            model_name="projectbillinghardcap",
            name="end_date",
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="projectbillinghardcap",
            name="start_date",
            field=models.DateField(blank=True, null=True),
        ),
    ]



# Generated by Django 4.2.23 on 2025-11-19 17:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("cap_discount", "0005_cap_multi_core_facilities"),
    ]

    operations = [
        migrations.AddField(
            model_name="capdiscount",
            name="creation_time",
            field=models.DateTimeField(null=True),
        ),
        migrations.AlterField(
            model_name="capdiscount",
            name="creation_time",
            field=models.DateTimeField(auto_now_add=True, null=True),
        ),
        migrations.AddField(
            model_name="capdiscountconfiguration",
            name="creation_time",
            field=models.DateTimeField(null=True),
        ),
        migrations.AlterField(
            model_name="capdiscountconfiguration",
            name="creation_time",
            field=models.DateTimeField(auto_now_add=True, null=True),
        ),
    ]



from NEMO.serializers import ModelSerializer
from NEMO.utilities import export_format_datetime
from NEMO.views.api import ModelViewSet, boolean_filters, date_filters, key_filters, number_filters
from drf_excel.mixins import XLSXFileMixin
from rest_flex_fields import FlexFieldsModelSerializer
from rest_framework import mixins
from rest_framework.fields import CharField
from rest_framework.viewsets import ReadOnlyModelViewSet

from NEMO_billing.rates.models import Rate, RateCategory, RateTime, RateType, DailySchedule


class RateCategorySerializer(ModelSerializer):
    class Meta:
        model = RateCategory
        fields = "__all__"


class RateTypeSerializer(ModelSerializer):
    class Meta:
        model = RateType
        fields = "__all__"


class RateTimeSerializer(ModelSerializer):
    class Meta:
        model = RateTime
        fields = "__all__"


class RateTimeDailyScheduleSerializer(ModelSerializer):
    class Meta:
        model = DailySchedule
        fields = "__all__"


class RateSerializer(FlexFieldsModelSerializer, ModelSerializer):
    class Meta:
        model = Rate
        fields = "__all__"
        expandable_fields = {
            "type": "NEMO_billing.rates.api.RateTypeSerializer",
            "tool": "NEMO.serializers.ToolSerializer",
            "area": "NEMO.serializers.AreaSerializer",
            "consumable": "NEMO.serializers.ConsumableSerializer",
        }

    type_name = CharField(source="type.get_type_display", read_only=True)
    category_name = CharField(source="category.name", default=None, read_only=True)
    tool_name = CharField(source="tool.name", default=None, read_only=True)
    area_name = CharField(source="area.name", default=None, read_only=True)
    consumable_name = CharField(source="consumable.name", default=None, read_only=True)


class RateCategoryViewSet(ModelViewSet):
    filename = "rate_categories"
    queryset = RateCategory.objects.all()
    serializer_class = RateCategorySerializer


class RateTypeViewSet(XLSXFileMixin, mixins.UpdateModelMixin, ReadOnlyModelViewSet):
    queryset = RateType.objects.all()
    serializer_class = RateTypeSerializer

    def get_filename(self, *args, **kwargs):
        return f"rate_types-{export_format_datetime()}.xlsx"


class RateTimeViewSet(ModelViewSet):
    filename = "rate_times"
    queryset = RateTime.objects.all()
    serializer_class = RateTimeSerializer


class RateTimeDailyScheduleViewSet(ModelViewSet):
    filename = "rate_time_daily_schedules"
    queryset = DailySchedule.objects.all()
    serializer_class = RateTimeDailyScheduleSerializer


class RateViewSet(ModelViewSet):
    filename = "rates"
    queryset = Rate.objects.all()
    serializer_class = RateSerializer
    filterset_fields = {
        "type": key_filters,
        "category": key_filters,
        "id": key_filters,
        "tool": key_filters,
        "area": key_filters,
        "consumable": key_filters,
        "amount": number_filters,
        "minimum_charge": number_filters,
        "service_fee": number_filters,
        "effective_date": date_filters,
        "flat": boolean_filters,
        "daily": boolean_filters,
        "daily_split_multi_day_charges": boolean_filters,
        "deleted": boolean_filters,
    }



# Generated by Django 3.2.18 on 2023-05-09 23:55

import django.core.validators
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("cap_discount", "0002_version_2_0_0"),
    ]

    def forwards_migrate_cap_discount(apps, schema_editor):
        CAPDiscountConfiguration = apps.get_model("cap_discount", "CAPDiscountConfiguration")
        CAPDiscount = apps.get_model("cap_discount", "CAPDiscount")
        CAPDiscountTier = apps.get_model("cap_discount", "CAPDiscountTier")
        for config in CAPDiscountConfiguration.objects.all():
            config.capdiscounttier_set.set(
                [
                    CAPDiscountTier.objects.create(
                        cap_discount_configuration=config, amount=config.amount, discount=config.discount
                    )
                ]
            )
        for discount in CAPDiscount.objects.all():
            discount.capdiscounttier_set.set(
                [
                    CAPDiscountTier.objects.create(
                        cap_discount=discount, amount=discount.amount, discount=discount.discount
                    )
                ]
            )

    def reverse_migrate_cap_discount(apps, schema_editor):
        CAPDiscountConfiguration = apps.get_model("cap_discount", "CAPDiscountConfiguration")
        CAPDiscount = apps.get_model("cap_discount", "CAPDiscount")
        for config in CAPDiscountConfiguration.objects.all():
            first_tier = config.capdiscounttier_set.first()
            config.amount = first_tier.amount
            config.discount = first_tier.discount
            config.save(update_fields=["amount", "discount"])
        for discount in CAPDiscount.objects.all():
            first_tier = discount.capdiscounttier_set.first()
            discount.amount = first_tier.amount
            discount.discount = first_tier.discount
            discount.save(update_fields=["amount", "discount"])

    operations = [
        migrations.CreateModel(
            name="CAPDiscountTier",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "amount",
                    models.DecimalField(
                        decimal_places=2,
                        help_text="Amount after which the discount will be applied",
                        max_digits=14,
                        validators=[django.core.validators.MinValueValidator(1)],
                    ),
                ),
                (
                    "discount",
                    models.DecimalField(
                        decimal_places=3,
                        help_text="Discount in percent. Ex 20.5%",
                        max_digits=6,
                        validators=[
                            django.core.validators.MinValueValidator(0),
                            django.core.validators.MaxValueValidator(100),
                        ],
                    ),
                ),
                (
                    "cap_discount",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="cap_discount.capdiscount",
                    ),
                ),
                (
                    "cap_discount_configuration",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="cap_discount.capdiscountconfiguration",
                    ),
                ),
            ],
            options={
                "ordering": ["amount"],
            },
        ),
        migrations.RunPython(forwards_migrate_cap_discount, reverse_migrate_cap_discount),
        migrations.AlterField(
            model_name="capdiscount",
            name="amount",
            field=models.DecimalField(
                default=0,
                decimal_places=2,
                help_text="Amount after which the discount will be applied",
                max_digits=14,
                validators=[django.core.validators.MinValueValidator(1)],
            ),
            preserve_default=True,
        ),
        migrations.RemoveField(
            model_name="capdiscount",
            name="amount",
        ),
        migrations.AlterField(
            model_name="capdiscount",
            name="discount",
            field=models.DecimalField(
                default=0,
                decimal_places=3,
                help_text="Discount in percent. Ex 20.5%",
                max_digits=6,
                validators=[django.core.validators.MinValueValidator(1), django.core.validators.MaxValueValidator(100)],
            ),
        ),
        migrations.RemoveField(
            model_name="capdiscount",
            name="discount",
        ),
        migrations.AlterField(
            model_name="capdiscountconfiguration",
            name="amount",
            field=models.DecimalField(
                default=0,
                decimal_places=2,
                help_text="Amount after which the discount will be applied",
                max_digits=14,
                validators=[django.core.validators.MinValueValidator(1)],
            ),
        ),
        migrations.RemoveField(
            model_name="capdiscountconfiguration",
            name="amount",
        ),
        migrations.AlterField(
            model_name="capdiscountconfiguration",
            name="discount",
            field=models.DecimalField(
                default=0,
                decimal_places=3,
                help_text="Discount in percent. Ex 20.5%",
                max_digits=6,
                validators=[django.core.validators.MinValueValidator(1), django.core.validators.MaxValueValidator(100)],
            ),
            preserve_default=True,
        ),
        migrations.RemoveField(
            model_name="capdiscountconfiguration",
            name="discount",
        ),
    ]



# Generated by Django 3.2.15 on 2022-11-12 18:51

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("invoices", "0008_version_1_10_0"),
    ]

    operations = [
        migrations.AddField(
            model_name="invoicedetailitem",
            name="content_type",
            field=models.ForeignKey(
                blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to="contenttypes.contenttype"
            ),
        ),
        migrations.AddField(
            model_name="invoicedetailitem",
            name="object_id",
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
        migrations.AddIndex(
            model_name="invoicedetailitem",
            index=models.Index(fields=["content_type", "object_id"], name="invoices_in_content_622541_idx"),
        ),
        migrations.AddField(
            model_name="invoiceconfiguration",
            name="hide_zero_charge",
            field=models.BooleanField(default=True, help_text="Hide charges with an amount equal to 0."),
        ),
    ]



# Generated by Django 2.2.13 on 2021-03-02 20:17

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("NEMO", "0027_version_3_8_0"),
    ]

    operations = [
        migrations.CreateModel(
            name="CoreFacility",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(help_text="The name of this core facility.", max_length=200, unique=True)),
            ],
            options={
                "verbose_name_plural": "Core facilities",
                "ordering": ["name"],
            },
        ),
        migrations.CreateModel(
            name="CustomCharge",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(help_text="The name of this custom charge.", max_length=255)),
                (
                    "additional_details",
                    models.CharField(
                        blank=True, help_text="Additional details for this charge.", max_length=255, null=True
                    ),
                ),
                ("date", models.DateTimeField(help_text="The date of the custom charge.")),
                (
                    "amount",
                    models.DecimalField(
                        decimal_places=2,
                        max_digits=8,
                        help_text="The amount of the charge. Use a negative amount for adjustments.",
                    ),
                ),
                (
                    "core_facility",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="NEMO_billing.CoreFacility",
                    ),
                ),
                (
                    "creator",
                    models.ForeignKey(
                        help_text="The person who created this charge.",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="custom_charge_creator",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "customer",
                    models.ForeignKey(
                        help_text="The customer to charge.",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="custom_charge_customer",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "project",
                    models.ForeignKey(
                        help_text="The project to bill for this charge.",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="NEMO.Project",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="CoreRelationship",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "area",
                    models.OneToOneField(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="core_rel",
                        to="NEMO.Area",
                    ),
                ),
                (
                    "consumable",
                    models.OneToOneField(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="core_rel",
                        to="NEMO.Consumable",
                    ),
                ),
                (
                    "core_facility",
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="NEMO_billing.CoreFacility"),
                ),
                (
                    "tool",
                    models.OneToOneField(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="core_rel",
                        to="NEMO.Tool",
                    ),
                ),
            ],
        ),
    ]
